#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.

__all__ = ['UpdateAclResponse', 'ListAclResponse']

from .acl_config import AclConfig
from .logresponse import LogResponse


class UpdateAclResponse(LogResponse):
    """ The response of the update_projecta_acl/update_logstore_acl API from log.
    
    :type header: dict
    :param header: UpdateAclResponse HTTP response header
    """

    def __init__(self, header, resp=''):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('UpdateAclResponse:')
        print('headers:', self.get_all_headers())


class ListAclResponse(LogResponse):
    """ The response of the get_project_acl/get_logstore_acl API from log.
    
    :type header: dict
    :param header: GetAclResponse HTTP response header
    """

    def __init__(self, resp, header):
        LogResponse.__init__(self, header, resp)
        self.count = int(resp["count"])
        self.total = int(resp["total"])
        self.acl_list = []
        for acl in resp.get("acls", []):
            acl_config = AclConfig(None, None)
            acl_config.from_json(acl)
            self.acl_list.append(acl_config)

    def get_acl_count(self):
        return self.count

    def get_acl_list(self):
        return self.acl_list

    def log_print(self):
        print('GetLogStoreResponse:')
        print('headers:', self.get_all_headers())
        print('acl_count:', str(self.count))
        print('acl_total:', str(self.total))
        print('acl_list:')
        for acl in self.acl_list:
            print(acl.to_json())



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.

from .logrequest import LogRequest


class ListTopicsRequest(LogRequest):
    """ The request used to get topics of a query from log. 
    
    :type project: string
    :param project: project name
    
    :type logstore: string
    :param logstore: logstore name
    
    :type token: string
    :param token: the start token to list topics
    
    :type line: int
    :param line: max topic counts to return
    """

    def __init__(self, project=None, logstore=None, token=None, line=None):
        LogRequest.__init__(self, project)
        self.logstore = logstore
        self.token = token
        self.line = line

    def get_logstore(self):
        """ Get logstore name
        
        :return: string, logstore name.
        """
        return self.logstore if self.logstore else ''

    def set_logstore(self, logstore):
        """ Set logstore name
        
        :type logstore: string
        :param logstore: logstore name
        """
        self.logstore = logstore

    def get_token(self):
        """ Get start token to list topics
        
        :return: string, start token to list topics
        """
        return self.token

    def set_token(self, token):
        """ Set start token to list topics
        
        :type token: string
        :param token: start token to list topics
        """
        self.token = token

    def get_line(self):
        """ Get max topic counts to return
        
        :return: int, max topic counts to return
        """
        return self.line

    def set_line(self, line):
        """ Set max topic counts to return
        
        :type line: int
        :param line: max topic counts to return
        """
        self.line = line



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.

from .util import Util
from .logresponse import LogResponse

__all__ = ['ListShardResponse', 'DeleteShardResponse']


class ListShardResponse(LogResponse):
    """ The response of the list_shard API from log.
    
    :type header: dict
    :param header: ListShardResponse HTTP response header

    :type resp: dict
    :param resp: the HTTP response body
    """

    def __init__(self, resp, header):
        LogResponse.__init__(self, header, resp)
        self._shards = Util.convert_unicode_to_str(resp)

    def get_shards_info(self):
        return self._shards

    def log_print(self):
        print('ListShardResponse:')
        print('headers:', self.get_all_headers())
        print("res:", self._shards)

    @property
    def shards(self):
        return self._shards

    @property
    def count(self):
        return len(self._shards)


class DeleteShardResponse(LogResponse):
    """ The response of the create_logstore API from log.
    
    :type header: dict
    :param header: DeleteShardResponse HTTP response header
    """

    def __init__(self, header, resp=''):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('DeleteShardResponse:')
        print('headers:', self.get_all_headers())



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.


class LogRequest(object):
    """ The base request of all log request. 
    
    :type project: string
    :param project: project name
    """

    def __init__(self, project):
        self.project = project

    def get_project(self):
        """ Get project name
        
        :return: string, project name.
        """
        return self.project if self.project else ''

    def set_project(self, project):
        """ Set project name
        
        :type project: string
        :param project: project name
        """
        self.project = project



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.
import json

from .logresponse import LogResponse

__all__ = ['ListIngestionResponse', 'CreateIngestionResponse', 'UpdateIngestionResponse', 'DeleteIngestionResponse', 'GetIngestionResponse',
           'StartIngestionResponse', 'StopIngestionResponse']


class ListIngestionResponse(LogResponse):
    """ The response of the ListIngestion API from log.

    :type resp: dict
    :param resp: ListIngestionResponse HTTP response body

    :type header: dict
    :param header: ListIngestionResponse HTTP response header
    """

    def __init__(self, resp, header):
        LogResponse.__init__(self, header, resp)
        self.count = resp['count']
        self.total = resp['total']
        self.ingestions = resp.get('results', [])

    def get_count(self):
        """ Get total count of ingestions from the response

        :return: int, the number of total ingestions from the response
        """
        return self.count

    def get_total(self):
        return self.total

    def get_ingestions(self):
        """ Get all the ingestions from the response

        :return: list, all ingestions
        """
        return self.ingestions

    def log_print(self):
        print('ListIngestionResponse:')
        print('headers:', self.get_all_headers())
        print('count:', self.count)
        print('total:', self.total)
        print('ingestions:', self.ingestions)


class CreateIngestionResponse(LogResponse):
    """ The response of the CreateIngestion API from log.

    :type resp: dict
    :param resp: CreateIngestionResponse HTTP response body

    :type header: dict
    :param header: CreateIngestionResponse HTTP response header
    """

    def __init__(self, header, resp=''):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('CreateIngestionResponse:')
        print('headers:', self.get_all_headers())


class UpdateIngestionResponse(LogResponse):
    """ The response of the UpdateIngestion API from log.

    :type resp: dict
    :param resp: UpdateIngestionResponse HTTP response body

    :type header: dict
    :param header: UpdateIngestionResponse HTTP response header
    """

    def __init__(self, header, resp=''):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('UpdateIngestionResponse:')
        print('headers:', self.get_all_headers())


class DeleteIngestionResponse(LogResponse):
    """ The response of the DeleteIngestion API from log.

    :type resp: dict
    :param resp: DeleteIngestionResponse HTTP response body

    :type header: dict
    :param header: DeleteIngestionResponse HTTP response header
    """

    def __init__(self, header, resp=''):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('DeleteIngestionResponse:')
        print('headers:', self.get_all_headers())


class GetIngestionResponse(LogResponse):
    """ The response of the GetIngestion API from log.

    :type resp: dict
    :param resp: GetIngestionResponse HTTP response body

    :type header: dict
    :param header: GetIngestionResponse HTTP response header
    """

    def __init__(self, resp, header):
        LogResponse.__init__(self, header, resp)
        self.ingestion_config = resp

    def get_ingestion(self):
        """ Get ingestion from the response

        :return: dict, ingestion
        """
        return self.ingestion_config

    def log_print(self):
        print('GetIngestionResponse:')
        print('headers:', self.get_all_headers())
        print('ingestion:', json.dumps(self.ingestion_config))


class StartIngestionResponse(LogResponse):
    """ The response of the StartIngestion API from log.

    :type resp: dict
    :param resp: StartIngestionResponse HTTP response body

    :type header: dict
    :param header: StartIngestionResponse HTTP response header
    """

    def __init__(self, header, resp=''):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('StartIngestionResponse:')
        print('headers:', self.get_all_headers())


class StopIngestionResponse(LogResponse):
    """ The response of the StopIngestion API from log.

    :type resp: dict
    :param resp: StopIngestionResponse HTTP response body

    :type header: dict
    :param header: StopIngestionResponse HTTP response header
    """

    def __init__(self, header, resp=''):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('StopIngestionResponse:')
        print('headers:', self.get_all_headers())



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.

from .logresponse import LogResponse
from .util import Util


class ListTopicsResponse(LogResponse):
    """ The response of the ListTopic API from log.

    :type resp: dict
    :param resp: ListTopicsResponse HTTP response body
    
    :type header: dict
    :param header: ListTopicsResponse HTTP response header
    """

    def __init__(self, resp, header):
        LogResponse.__init__(self, header, resp)
        self.count = Util.h_v_t(header, 'x-log-count')
        self.nextToken = Util.h_v_t(header, 'x-log-nexttoken')
        self.topics = resp

    def get_count(self):
        """ Get the number of all the topics from the response
        
        :return: int, the number of all the topics from the response
        """
        return self.count

    def get_topics(self):
        """ Get all the topics from the response
        
        :return: list, topic list
        """
        return self.topics

    def get_next_token(self):
        """ Return the next token from the response. If there is no more topic
        to list, it will return None
        
        :return: string, next token used to list more topics
        """
        return self.nextToken

    def log_print(self):
        print('ListTopicsResponse:')
        print('headers:', self.get_all_headers())
        print('count:', self.count)
        print('nextToken:', self.nextToken)
        print('topics:', self.topics)



import inspect
import json
import logging
try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

import jmespath
import re
import six
from jmespath.exceptions import ParseError

from .trans_base import trans_comp_check_mdoe_base
from ..etl_util import get_re_full_match
from ..exceptions import SettingError

__all__ = ['trans_comp_json']

logger = logging.getLogger(__name__)


def trans_comp_json(*args, **kwargs):
    if (args and isinstance(args[0], dict)) or ('event' in kwargs and isinstance(kwargs['event'], dict)):
        # event case
        trans = json_transformer()
        return trans(*args, **kwargs)
    else:
        return json_transformer(*args, **kwargs)


class json_transformer(trans_comp_check_mdoe_base):
    DEFAULT_SEP = u'.'
    DEFAULT_FMT = "simple"
    DEFAULT_DEPTH = 100
    DEFAULT_INCLUDE_NODE = trans_comp_check_mdoe_base.DEFAULT_KEYWORD_PTN
    DEFAULT_EXCLUDE_NODE = u''
    DEFAULT_INCLUDE_PATH = u''
    DEFAULT_EXCLUDE_PATH = u''

    DEFAULT_FMT_ARRAY = u"{parent_rlist[0]}_{index}"  # could also be custom formatting string using up to five placehodler: parent_list, parent_list, current, sep, prefix, suffix
    FMT_MAP = {
        "simple": lambda prefix, current, suffix, *args, **kwargs: u"{prefix}{current}{suffix}".format(prefix=prefix,
                                                                                                      current=current,
                                                                                                      suffix=suffix),
        "full": lambda parent_list, sep, prefix, current, suffix, *args,
                       **kwargs: u"{parent_list_str}{sep}{prefix}{current}{suffix}".format(
            parent_list_str=sep.join(parent_list), current=current, sep=sep, prefix=prefix, suffix=suffix),
        "parent": lambda parent_list, sep, prefix, current, suffix, *args,
                         **kwargs: u"{parent}{sep}{prefix}{current}{suffix}".format(parent=parent_list[-1],
                                                                                   current=current, sep=sep,
                                                                                   prefix=prefix, suffix=suffix),
        "root": lambda parent_list, sep, prefix, current, suffix, *args,
                       **kwargs: u"{parent_list[0]}{sep}{prefix}{current}{suffix}".format(parent_list=parent_list,
                                                                                         current=current, sep=sep,
                                                                                         prefix=prefix, suffix=suffix)
        # could also be custom formatting string using up to five placehodler: parent_list, parent_list, current, sep, prefix, suffix
        # could also be formatting function accepting the 3 parameters: parrent_list, current key, current value
        #    Note: the functoin must result k, v tuple, if returning None, meaning skip this k-v
    }

    def __init__(self, jmes=None, jmes_ignore_none=None, output=None,
                 expand=None, depth=None, include_node=None, exclude_node=None,
                 include_path=None, exclude_path=None,
                 fmt=None, sep=None, prefix=None, suffix=None,
                 expand_array=None, fmt_array=None, mode=None
                 ):
        """
        :param jmes:  jmes filter to select or generate new field
        :param jmes_ignore_none: if jmes filter is null, ignore it (default). Or else consider it as "". default is
        :param output: put the value parsed from jmes filter to this field
        :param expand: If jmes filter is configure, expand the result or not (Default is False in this case), If jmes is not configured, directly expand the field passed or not (Default is True in this case).
        :param depth: depth to scan, 1 means first layer, default is 100.
        :param include_node: keys to expand and include. regex string. using '|' for multiple ones. default is all.
        :param exclude_node: keys to skip, regex string. using '|' for multiple ones. default is nothing.
        :param include_path: path to expand and include. regex string to match from begin. using '|' for multiple ones. default is all. e.g. r"data\.k1", all sub-keys in data.k1 will be included.
        :param exclude_path: path to skip, regex string to match from begin. using '|' for multiple ones. default is nothing. . e.g. r"data\.k2", all sub-keys in data.k2 will be excluded.
        :param fmt: during expansion, how to format the key name, there're several types or customized as described in FMT_MAP
        :param sep: sep used in formatting during expansion
        :param prefix: prefix used in formatting during expansion
        :param suffix: suffix used in formatting during expansion
        :param expand_array: if expand array or just render it. default is True. item in array will be with name index, e.g. [1,2,3] will be considered as {"0": 1, "1": 2, "2": 3}
        :param fmt_array: format string for key name of each array element, default is "{parent_rlist[0]}_{index}", can be custom formatting string using placehodler: parent_list, parent_list, current
        """
        super(json_transformer, self).__init__(mode=mode)

        self.expand = expand
        if expand is None:
            # when jmes is not configured or configure but no output configured
            self.expand = not jmes or not output

        # self.level = level or 1
        self.jmes = self._u(jmes or "")
        self.prefix = self._u("" if prefix is None else prefix)
        self.suffix = self._u("" if suffix is None else suffix)
        self.sep = self._u(self.DEFAULT_SEP if sep is None else sep)
        self.output = self._u(output or "")
        self.jmes_filter = None
        self.jmes_ignore_none = True if jmes_ignore_none is None else jmes_ignore_none
        if jmes:
            try:
                self.jmes_filter = jmespath.compile(jmes)
            except jmespath.exceptions.ParseError as ex:
                raise SettingError(ex=ex, msg="Invalid JMES filter setting", settings=jmes)
        elif self.output:
            logger.warning(u"json_transformer: parameter output '{0}' will be ignored as there's no filter is selected."
                           .format(output))

        self.depth = min((depth or self.DEFAULT_DEPTH), self.DEFAULT_DEPTH)
        self.include_node = self._u(include_node or self.DEFAULT_INCLUDE_NODE)
        self.exclude_node = self._u(exclude_node or self.DEFAULT_EXCLUDE_NODE)
        self.include_path = self._u(include_path or self.DEFAULT_INCLUDE_PATH)
        self.exclude_path = self._u(exclude_path or self.DEFAULT_EXCLUDE_PATH)
        self.fmt = self._u(fmt or self.DEFAULT_FMT)

        try:
            self.include_node_match = get_re_full_match(self.include_node)
            self.exclude_node_match = get_re_full_match(self.exclude_node)
            self.include_path_match = re.compile(self.include_path).match  # use match instead of full match
            self.exclude_path_match = re.compile(self.exclude_path).match  # use match instead of full match
        except Exception as ex:
            raise SettingError(ex=ex, msg="Invalid regex string for include/exclude")

        self.expand_array = True if expand_array is None else expand_array
        self.format_array = self._u(fmt_array or self.DEFAULT_FMT_ARRAY)

    def _skip_keys(self, key, parent_list):
        if (self.include_node and not self.include_node_match(key)) or (
                self.exclude_node and self.exclude_node_match(key)):
            logger.info(u"json_transformer: 'key' {0} is not in include keys '{1}' or in exclude keys '{2}', skip it."
                        .format(key, self.include_node, self.exclude_node))
            return True

        if self.include_path or self.exclude_path:
            path = '.'.join(parent_list) + '.' + key
            if (self.include_path and not self.include_path_match(path)) or (
                    self.exclude_path and self.exclude_path_match(path)):
                logger.info(
                    u"json_transformer: path '{0}' is not in include path '{1}' or in exclude path '{2}', skip it."
                    .format(path, self.include_path, self.exclude_path))
                return True

        return False

    def format_add_kv(self, event, fmt, current, value, parent_list, parent_rlist, sep, prefix, suffix):
        if self._skip_keys(current, parent_list):
            logger.info(u"json_transformer: 'key' {0} is not in include keys '{1}' or in exclude keys '{2}', skip it."
                        .format(current, self.include_node, self.exclude_node))
            return

        ret = None
        if isinstance(fmt, (six.text_type, six.binary_type)):
            fmt = json_transformer.FMT_MAP.get(fmt.strip().lower(), fmt)
            try:
                if isinstance(fmt, (six.text_type, six.binary_type)):
                    ret = fmt.format(parent_list=parent_list, parent_rlist=parent_rlist, current=current, sep=sep,
                                     prefix=prefix, suffix=suffix), \
                          json_transformer._n(value)
                else:
                    # callable formatting function
                    ret = fmt(parent_list=parent_list, parent_rlist=parent_rlist, current=current, sep=sep,
                              prefix=prefix, suffix=suffix), \
                          json_transformer._n(value)
            except Exception as ex:
                logger.info(u"json_transformer: fail to format with settings: '{0}'".format((fmt, current, value,
                                                                                            parent_list, sep, prefix,
                                                                                            suffix)))
        elif inspect.isfunction(fmt):
            try:
                ret = fmt(parent_list, current, value)
            except Exception as ex:
                logger.info(u"json_transformer: fail to call formatting string: {0} wuth parameters: {1}"
                            .format(fmt, (parent_list, current, value)))

        if ret and len(ret) == 2:
            k, v = ret
            event[json_transformer._n(k)] = json_transformer._n(v)
        else:
            logger.info(u"json_transformer: unexpected format result: {0}, fmt: '{1}', k: '{2}', v: '{3}', skip it"
                        .format(ret, fmt, current, value))

    def _expand_json(self, event, key, value, parent_list, parent_rlist, depth, sep, prefix, suffix):
        # check if need to format it directly
        if len(parent_list) > depth \
                or (not isinstance(value, (list, tuple, dict))) \
                or (isinstance(value, (list, tuple)) and not self.expand_array):
            # 1. depth hit, 2. basic type, 3. array but not expand
            logger.info(u"json_transformer: hit stop parsing, key: '{0}', value: '{1}', parent: '{2}', depth: '{3}'"
                        .format(key, value, parent_list, depth))
            self.format_add_kv(event, self.fmt, self._n(key), self._n(value), parent_list, parent_rlist, sep, prefix,
                               suffix)
            return None

        # convert array to dict
        if isinstance(value, (list, tuple)):
            value = dict(
                (self.format_array.format(parent_list=parent_list, parent_rlist=parent_rlist, index=i), v) for i, v in
                enumerate(value))

        if isinstance(value, dict):
            for k, v in six.iteritems(value):
                if isinstance(v, (dict, tuple, list)):
                    # recursively parse it
                    self._expand_json(event, k, v, parent_list + (k,), (k,) + parent_rlist, depth, sep, prefix, suffix)
                else:
                    self.format_add_kv(event, self.fmt, self._n(k), self._n(v), parent_list, parent_rlist, sep, prefix,
                                       suffix)

        else:
            logger.info(u"json_transformer: skip unsupported message '{0}' of type '{1}' when expanding"
                        .format(value, type(value)))

    def _process_message(self, key, value):
        new_event = {}
        if isinstance(value, (six.binary_type, six.text_type)):
            try:
                value = json.loads(value)
            except Exception as ex:
                logger.info(u"json_transformer: fail to load event into json object: {0}, error: {1}".format(value, ex))
                return None

        if self.jmes_filter:
            try:
                value = self.jmes_filter.search(value)
                if value is None and self.jmes_ignore_none:
                    logger.info(u"split_event_transformer: value {0} get null from jmes settings {1}, skip it".
                                format(value, self.jmes))
                    return None
            except Exception as ex:
                logger.info(u"split_event_transformer: value {0} with invalid jmes settings {1}, skip it".
                            format(value, self.jmes))
                return None

            if self.output:
                new_event[self.output] = self._n(value)

        # if need to expand
        if self.expand:
            self._expand_json(new_event, key, value, (key,), (key,), self.depth, self.sep, self.prefix, self.suffix)

        return new_event

    def __call__(self, event, inpt):
        inpt = self._u(inpt)

        # simple dict mode
        if isinstance(inpt, (six.binary_type, six.text_type)):
            inpt = [inpt]

        if isinstance(inpt, Iterable):
            for i in inpt:
                if not isinstance(i, (six.binary_type, six.text_type)):
                    logger.error(u'trans_comp_lookup: type of input field "{0}" is unknown'.format(i))
                    continue

                if i not in event:
                    logger.info(u'trans_comp_lookup: event "{0}" does not contain field "{1}"'.format(event, i))
                    continue

                # get input value
                new_event = self._process_message(i, event[i])
                if new_event and isinstance(new_event, dict):
                    #event.update(new_event)
                    self.sets(event, new_event)
                else:
                    logger.info(
                        u'trans_comp_lookup: event "{0}" does not extract value from field "{1}"'.format(event, i))
        else:
            logger.error(u"trans_comp_lookup: unknown type of input field {0}".format(inpt))

        return event



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.


import time
import logging
from aliyun.log import (
    LogException, LogClient, SimpleLogHandler, IndexLineConfig, IndexConfig,
)


_log_fields = [
    'asctime',
    'filename',
    'funcName',
    'levelname',
    'lineno',
    'module',
    'message',
    'process',
]
_migration_logstore = 'internal-es-migration-log'


def setup_logging(migration, endpoint, project, access_key_id, access_key):
    # setup internal logstore
    _setup_migration_logstore(endpoint, project, access_key_id, access_key)
    time.sleep(10)

    # set logging level for libs
    logging.getLogger('requests').setLevel(logging.WARNING)
    logging.getLogger('urllib3').setLevel(logging.WARNING)
    logging.getLogger('elasticsearch').setLevel(logging.WARNING)

    # add aliyunlog handler
    root = logging.getLogger()
    root.setLevel(logging.INFO)
    fields = 'level,process_id,thread_id,module,func_name'.split(',')
    handler = SimpleLogHandler(
        end_point=endpoint,
        access_key_id=access_key_id,
        access_key=access_key,
        project=project,
        log_store=_migration_logstore,
        fields=fields,
        extract_json=True,
    )
    handler.skip_message = False
    handler.built_in_root_field = 'logging'
    handler.log_tags = [
        ('__migration__', migration),
    ]
    root.addHandler(handler)


def _setup_migration_logstore(endpoint, project, access_key_id, access_key):
    log_client = LogClient(
        endpoint=endpoint,
        accessKeyId=access_key_id,
        accessKey=access_key,
    )
    try:
        log_client.create_logstore(
            project_name=project,
            logstore_name=_migration_logstore,
        )
    except LogException as exc:
        if exc.get_error_code() != "LogStoreAlreadyExist":
            raise
    try:
        tokens = [
            ',', ' ', "'", '"', ';', '=', '(', ')', '[', ']', '{', '}', '?',
            '@', '&', '<', '>', '/', ':', '\n', '\t', '\r',
        ]
        line_config = IndexLineConfig(token_list=tokens)
        config = IndexConfig(line_config=line_config)
        log_client.create_index(project, _migration_logstore, config)
    except LogException as exc:
        if exc.get_error_code() != "IndexAlreadyExist":
            raise



# -*- coding: utf-8 -*-

import time

from .exceptions import CheckPointException
from ..logexception import LogException
from multiprocessing import RLock

class ConsumerCheckpointTracker(object):

    def __init__(self, loghub_client_adapter, consumer_name, shard_id):
        self.consumer_group_client = loghub_client_adapter
        self.consumer_name = consumer_name
        self.shard_id = shard_id
        self.last_check_time = time.time()
        self.cursor = ''
        self.temp_check_point = ''
        self.last_persistent_checkpoint = ''
        self.default_flush_check_point_interval = 60
        self.lock = RLock()

    def set_cursor(self, cursor):
        self.cursor = cursor

    def get_cursor(self):
        return self.cursor

    def save_check_point(self, persistent, cursor=None):
        if cursor is not None:
            self.temp_check_point = cursor
        else:
            self.temp_check_point = self.cursor
        if persistent:
            self.flush_check_point()

    def set_memory_check_point(self, cursor):
        self.temp_check_point = cursor

    def set_persistent_check_point(self, cursor):
        self.last_persistent_checkpoint = cursor

    def flush_check_point(self):
        with self.lock:
            if self.temp_check_point != '' and self.temp_check_point != self.last_persistent_checkpoint:
                try:
                    self.consumer_group_client.update_check_point(
                        self.shard_id, self.consumer_name, self.temp_check_point)
                    self.last_persistent_checkpoint = self.temp_check_point
                except LogException as e:
                    raise CheckPointException("Failed to persistent the cursor to outside system, " +
                                              self.consumer_name + ", " + str(self.shard_id)
                                              + ", " + self.temp_check_point, e)

    def flush_check(self):
        current_time = time.time()
        if current_time > self.last_check_time + self.default_flush_check_point_interval:
            try:
                self.flush_check_point()
            except CheckPointException as e:
                print(e)
            self.last_check_time = current_time

    def get_check_point(self):
        return self.temp_check_point




# encoding: utf8

from IPython.core.magic import (Magics, magics_class, line_magic,
                                cell_magic, line_cell_magic)
import pandas as pd
from IPython.display import display, clear_output
import re, time, threading, datetime
from pandas import DataFrame
from aliyun.log import LogClient, LogException
from concurrent.futures import ThreadPoolExecutor as PoolExecutor, as_completed
import multiprocessing
import six
import six.moves.configparser as configparser
import os
import sys


def can_use_widgets():
    """ Expanded from from http://stackoverflow.com/a/34092072/1958900
    """
    if 'IPython' not in sys.modules:
        # IPython hasn't been imported, definitely not
        return False
    from IPython import get_ipython

    # check for `kernel` attribute on the IPython instance
    if getattr(get_ipython(), 'kernel', None) is None:
        return False

    try:
        import ipywidgets as ipy
        import traitlets
    except ImportError:
        return False

    if int(ipy.__version__.split('.')[0]) < 6:
        print('WARNING: widgets require ipywidgets 6.0 or later')
        return False

    return True

__CAN_USE_WIDGET__ = can_use_widgets()


CLI_CONFIG_FILENAME = "%s/.aliyunlogcli" % os.path.expanduser('~')
MAGIC_SECTION = "__jupyter_magic__"
DEFAULT_DF_NAME = 'log_df'
DEFAULT_TMP_DF_NAME = 'log_df_part'

result = None
detail = None


def _load_config():
    global g_default_region, g_default_ak_id,  g_default_ak_key, g_default_project, g_default_logstore

    def _get_section_option(config, section_name, option_name, default=None):
        if six.PY3:
            return config.get(section_name, option_name, fallback=default)
        else:
            return config.get(section_name, option_name) if config.has_option(section_name, option_name) else default

    config = configparser.ConfigParser()
    config.read(CLI_CONFIG_FILENAME)
    g_default_region = _get_section_option(config, MAGIC_SECTION, "region-endpoint", "")
    g_default_ak_id = _get_section_option(config, MAGIC_SECTION, "access-id", "")
    g_default_ak_key = _get_section_option(config, MAGIC_SECTION, "access-key", "")
    g_default_project = _get_section_option(config, MAGIC_SECTION, "project", "")
    g_default_logstore = _get_section_option(config, MAGIC_SECTION, "logstore", "")


def _save_config(region, ak_id, ak_key, project, logstore):
    global g_default_region, g_default_ak_id, g_default_ak_key, g_default_project, g_default_logstore

    config = configparser.ConfigParser()
    config.read(CLI_CONFIG_FILENAME)

    if not config.has_section(MAGIC_SECTION):
        config.add_section(MAGIC_SECTION)

    config.set(MAGIC_SECTION, "region-endpoint", region)
    config.set(MAGIC_SECTION, "access-id", ak_id)
    config.set(MAGIC_SECTION, "access-key", ak_key)
    config.set(MAGIC_SECTION, "project", project)
    config.set(MAGIC_SECTION, "logstore", logstore)

    # save to disk
    with open(CLI_CONFIG_FILENAME, 'w') as configfile:
        config.write(configfile)

    # save to memory
    g_default_region, g_default_ak_id, g_default_ak_key, g_default_project, g_default_logstore = region, ak_id, ak_key, project, logstore

_load_config()


def parse_timestamp(tm):
    return datetime.datetime.fromtimestamp(int(tm)).isoformat()


@magics_class
class MyMagics(Magics):
    logclient = None

    @staticmethod
    def pull_worker(client, project_name, logstore_name, from_time, to_time, shard_id):
        res = client.pull_log(project_name, logstore_name, shard_id, from_time, to_time)
        result = []
        next_cursor = 'as from_time configured'
        try:
            for data in res:
                result.extend(data.get_flatten_logs_json(decode_bytes=True))
                next_cursor = data.next_cursor
        except Exception as ex:
            print("dump log failed: task info {0} failed to copy data to target, next cursor: {1} detail: {2}".
                format(
                (project_name, logstore_name, shard_id, from_time, to_time),
                next_cursor, ex))

        return result

    @staticmethod
    def pull_log_all(client, project_name, logstore_name, from_time, to_time):
        cpu_count = multiprocessing.cpu_count() * 2

        shards = client.list_shards(project_name, logstore_name).get_shards_info()
        current_shards = [str(shard['shardID']) for shard in shards]
        target_shards = current_shards
        worker_size = min(cpu_count, len(target_shards))

        result = []
        with PoolExecutor(max_workers=worker_size) as pool:
            futures = [pool.submit(MyMagics.pull_worker, client, project_name, logstore_name, from_time, to_time,
                                   shard_id=shard)
                       for shard in target_shards]

            try:
                for future in as_completed(futures):
                    data = future.result()
                    result.extend(data)

                return True, result
            except KeyboardInterrupt as ex:
                clear_output()
                print(u"正在取消当前获取……")
                for future in futures:
                    if not future.done():
                        future.cancel()

        return False, result

    def client(self, reset=False):
        if self.logclient is None or reset:
            self.logclient = LogClient(g_default_region, g_default_ak_id, g_default_ak_key)
        return self.logclient

    def verify_sls_connection(self, region, ak_id, ak_key, project, logstore):
        logclient = LogClient(region, ak_id, ak_key)
        try:
            res = logclient.get_logstore(project, logstore)
            return True, res.body
        except LogException as ex:
            return False, str(ex)
        except Exception as ex:
            return False, str(ex)

    @staticmethod
    def _get_log_param(line, cell):
        to_time = time.time()
        from_time = to_time - 60*15
        if cell is None:
            query = line
        else:
            line = line.strip()
            ret = [line, '']
            if '~' in line:
                ret = line.split('~')
            elif '-' in line:
                ret = line.split('~')

            from_time = ret[0].strip() or from_time
            to_time = ret[1].strip() or to_time if len(ret) > 1 else to_time

            query = cell

        return from_time, to_time, query

    def log_imp(self, line, cell=None):
        from_time, to_time, query = self._get_log_param(line, cell)

        print(u"根据查询统计语句，从日志服务查询数据(时间范围：{0} ~ {1})，结果将保存到变量{2}中，请稍等……".format(from_time, to_time, DEFAULT_DF_NAME))
        res = self.client().get_log_all(g_default_project, g_default_logstore, from_time=from_time, to_time=to_time,
                                         query=query)
        is_complete = True
        logs = []
        try:
            for data in res:
                if not data.is_completed():
                    is_complete = False
                logs.extend(data.body)
        except Exception as ex:
            print(ex)
            return

        df1 = pd.DataFrame(logs)
        is_stat = re.match(r'.+\|\s+select\s.+|^\s*select\s+.+', query.lower(), re.I) is not None
        if is_stat:
            if "__time__" in df1:
                del df1["__time__"]
            if "__source__" in df1:
                del df1["__source__"]
        else:
            # change time to date time
            if "__time__" in df1:
                df1['__time__'] = pd.to_datetime(df1["__time__"].apply(parse_timestamp))
                df1.set_index('__time__', inplace=True)

        clear_output()
        if is_complete:
            print(u"变量名：{0}".format(DEFAULT_DF_NAME))
        elif is_stat:
            print(u"变量名：{0}，结果非完整精确结果。".format(DEFAULT_DF_NAME))
        else:
            print(u"变量名：{0}，部分结果非完整精确结果。".format(DEFAULT_DF_NAME))

        self.shell.user_ns[DEFAULT_DF_NAME] = df1
        return df1

    def fetch_log_imp(self, line):
        from_time, to_time, _ = self._get_log_param(line, "")

        print(u"从日志服务拉取数据(日志插入时间：{0} ~ {1})，结果将保存到变量{2}中，请稍等……".format(from_time, to_time, DEFAULT_DF_NAME))
        result, logs = self.pull_log_all(self.client(), g_default_project, g_default_logstore, from_time, to_time)

        df1 = pd.DataFrame(logs)

        # change time to date time
        if "__time__" in df1:
            df1['__time__'] = pd.to_datetime(df1["__time__"].apply(parse_timestamp))
            df1.set_index('__time__', inplace=True)

        clear_output()
        if not result:
            print(u"获取被取消，显示部分结果，变量名：{0}".format(DEFAULT_TMP_DF_NAME))
            self.shell.user_ns[DEFAULT_TMP_DF_NAME] = df1
        else:
            print(u"变量名：{0}".format(DEFAULT_DF_NAME))
            self.shell.user_ns[DEFAULT_DF_NAME] = df1
        return df1

    @line_cell_magic
    def log(self, line, cell=None):
        return self.log_imp(line, cell)

    @line_cell_magic
    def fetch(self, line):
        return self.fetch_log_imp(line)

    @line_magic
    def manage_log(self, line):
        line = line or ""
        if line or not __CAN_USE_WIDGET__:
            params = line.split(" ")
            if len(params) == 5:
                print(u"连接中...")
                endpoint, key_id, key_val, project, logstore = params
                result, detail = self.verify_sls_connection(endpoint, key_id, key_val, project, logstore)

                if result:
                    clear_output()
                    print(u"连接成功.")
                    _save_config(endpoint, key_id, key_val, project, logstore)
                    self.client(reset=True)
                else:
                    print(detail)
            else:
                print(u"参数错误，请使用GUI配置（无参）或遵循格式：%manage_log <endpoint> <ak_id> <ak_key> <project> <logstore>")

            return

        import ipywidgets as widgets
        w_1 = widgets.ToggleButtons( options=[u'基本配置', u"高级配置"] )

        w_endpoint = widgets.Text( description=u'服务入口', value=g_default_region)
        w_key_id = widgets.Password( description=u'秘钥ID', value=g_default_ak_id)
        w_key_val = widgets.Password( description=u'秘钥Key', value=g_default_ak_key)
        w_project = widgets.Text( description=u'默认项目', value=g_default_project)
        w_logstore = widgets.Text( description=u'默认日志库', value=g_default_logstore)
        w_confirm = widgets.Button(
            description=u'修改' if g_default_region else u'确认',
            button_style='info',
            icon='confirm'
        )
        w_result = widgets.Label(value='')
        hide_layout = widgets.Layout(height="0px")
        show_layout = widgets.Layout(height='auto')
        progress = widgets.FloatProgress(description="", value=0.0, min=0.0, max=1.0, layout=hide_layout)

        def work(progress):
            global result
            total = 100
            for i in range(total):
                time.sleep(0.2)
                if result is None:
                    progress.value = float(i+1)/total
                else:
                    progress.value = 100
                    progress.description = u"完成" if result else u"失败"
                    break
                
        def on_button_clicked(b):
            global result, detail
            progress.layout = show_layout
            progress.description = u"连接中..."
            progress.value = 0
            w_result.value = ""
               
            result = None
            detail = ""
            thread = threading.Thread(target=work, args=(progress,))
            thread.start()

            result, detail = self.verify_sls_connection(w_endpoint.value, w_key_id.value, w_key_val.value, w_project.value, w_logstore.value)

            if result:
                w_result.value = u"连接成功."
                _save_config(w_endpoint.value, w_key_id.value, w_key_val.value, w_project.value, w_logstore.value)
                self.client(reset=True)
            else:
                w_result.value = str(detail)

        w_confirm.on_click(on_button_clicked)

        p = widgets.VBox(children=[w_1, w_endpoint, w_key_id, w_key_val, w_project, w_logstore, w_confirm, progress, w_result])
        
        return p
    

def df_html(df1):
    if not __CAN_USE_WIDGET__:
        return df1._repr_html_()

    try:
        import odps
        if len(df1.columns) > 1:
            df2 = odps.DataFrame(df1)
            return df2._repr_html_()
        return df1._repr_html_()
    except Exception as ex:
        print(ex)
        return df1._repr_html_()


def load_ipython_extension(ipython):
    ipython.register_magics(MyMagics)
    html_formatter = ipython.display_formatter.formatters['text/html']
    html_formatter.for_type(DataFrame, df_html)



from .transform import *
from enum import Enum

builtin_macros = {
    'KEEP_EVENT_.*': keep_event,
    'DROP_EVENT_.*': drop_event,
    'KEEP_FIELDS_.*': keep_fields,
    'DROP_FIELDS_.*': drop_fields,
    'RENAME_FIELDS_.*': rename_fields,
    'ALIAS_.*': rename_fields,
    'DISPATCH_EVENT_.*': dispatch_event,
    'TRANSFORM_EVENT_.*': transform_event,
    'KV_FIELDS_.*': extract_kv_fields
}

__all__ = ['KEEP_EVENT_',
           'DROP_EVENT_',
           'KEEP_FIELDS_',
           'DROP_FIELDS_',
           'RENAME_FIELDS_',
           'KV_FIELDS_',
           'ALIAS_',
           'DISPATCH_EVENT_',
           'TRANSFORM_EVENT_']

for key in builtin_macros.keys():
    globals()[key[:-2]] = "Use this prefix to auto call the function: {0}".format(builtin_macros[key])


class TransFnType(Enum):
    EVENT_NEW = "EVENT_NEW",
    EVENT_UPDATE = 'EVENT_UPDATE',
    # TRANS_COMP = 'TRANS_COMP'
    UNKNOWN = 'UNKNOWN'


def check_fn_type_by_name(name):
    # if isinstance(fn, transform_base):
    #     return TransFnType.EVENT_NEW
    # elif isinstance(fn, trans_comp_base):
    #     return TransFnType.TRANSFORM_COMP
    # if not isinstance(fn, Callable):
    #     fn_name = getattr(fn, '__name__', getattr(fn, "func_name", ""))
    if name.startswith("sls_en_"):
        return TransFnType.EVENT_NEW
    elif name.startswith("sls_eu_"):
        return TransFnType.EVENT_UPDATE

    return TransFnType.UNKNOWN



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.

import time

from .util import Util
import copy
import warnings
from enum import Enum
from .logexception import LogException
import logging

__all__ = ['PluginConfigDetail', 'SeperatorFileConfigDetail', 'SimpleFileConfigDetail', 'FullRegFileConfigDetail',
           'JsonFileConfigDetail', 'ApsaraFileConfigDetail', 'SyslogConfigDetail',
           'LogtailConfigGenerator', 'CommonRegLogConfigDetail']

logger = logging.getLogger(__name__)


class LogtailInputType(Enum):
    FILE = 'file'
    SYSLOG = 'syslog'
    PLUGIN = 'plugin'


class LogtailType(Enum):
    JSON = 'json_log'
    FULL_REGEX = 'common_reg_log'
    SEPARATOR = 'delimiter_log'
    APSARA = "apsara_log"


class LogtailConfigDetail(object):
    """The common parts of logtail config
    :type config_name: string
    :param config_name: the config name

    :type logstore_name: string
    :param logstore_name: the logstore name for the config 

    :type endpoint: string
    :param endpoint: deprecated, set it as empty

    :type log_path: string
    :param log_path: the log file dir path

    :type file_pattern: string
    :param file_pattern: the log file name pattern, e.g *.LOG , access.log

    :type log_begin_regex: string
    :param log_begin_regex: the regular express to match the first line of a log

    :type topic_format: string
    :param topic_format: "none" or "group_topic"

    :type filter_keys: string list
    :param filter_keys: the keys used to filter logs, e.g ["key_1", "key_2"]

    :type filter_keys_reg: string list
    :param filter_keys_reg: the regex for filter_keys to filter the log, filter_keys_reg[i] is for filter_keys[i].
    The size of filter_keys_reg and filter_keys should be same.\
        If a log is matched only if the size of filter_keys is 0, or all the value of the related keys in filter_keys,
        match the regex set in filter_keys_reg

    :type logSample: string
    :param logSample: sample strings for the log, up to 1000 bytes

    :type log_type: string
    :param log_type: common_reg_log, delimiter_log, apsara_log etc.

    :type extended_items: dict
    :param extended_items: extended items in dict format e.g. enableRawLog etc. refer to wiki page.
    https://help.aliyun.com/document_detail/29042.html?spm=5176.doc28997.6.744.C583Jg

    """

    def __init__(self, config_name, logstore_name, endpoint, log_path, file_pattern, log_begin_regex, topic_format,
                 filter_keys, filter_keys_reg, logSample='', log_type='common_reg_log', **extended_items):
        logger.warning("aliyun.log.LogtailConfigDetail is deprecated and will be removed in future version.")

        self.config_name = config_name
        self.logstore_name = logstore_name
        self.endpoint = endpoint or ''
        self.log_path = log_path
        self.file_pattern = file_pattern

        self.log_begin_regex = log_begin_regex
        self.topic_format = topic_format
        self.filter_keys = filter_keys
        self.filter_keys_reg = filter_keys_reg
        self.create_time = int(time.time())
        self.last_modify_time = int(time.time())
        self.logSample = logSample
        self.log_type = log_type or 'common_reg_log'
        self.extended_items = extended_items

    def set_create_time(self, create_time):
        self.create_time = create_time

    def set_last_modify_time(self, last_modify_time):
        self.last_modify_time = last_modify_time

    @staticmethod
    def from_json(json_value):
        logger.warning("aliyun.log.LogtailConfigDetail is deprecated and will be removed in future version.")
        return LogtailConfigHelper.generate_logtail_config(json_value)


class CommonRegLogConfigDetail(LogtailConfigDetail):
    """The logtail config for common_reg_log

    :type config_name: string
    :param config_name: the config name

    :type logstore_name: string
    :param logstore_name: the logstore name for the config 

    :type endpoint: string
    :param endpoint: log data endpoint, deprecated, set it as empty

    :type log_path: string
    :param log_path: the log file dir path

    :type file_pattern: string
    :param file_pattern: the log file name pattern, e.g \*.LOG , access.log

    :type time_format: string
    :param time_format: the time format of the logs, e.g.  "%Y-%m-%d %M:%H:%S"

    :type log_begin_regex: string
    :param log_begin_regex: the regular express to match the first line of a log

    :type log_parse_regex: string
    :param log_parse_regex: the regular express to match a log, e.g (\d+-\d+\d+ \d+:\d+:\d+) (\S+) (.*)

    :type reg_keys: string list
    :param reg_keys: the key for every captured value in log_parse_reg, e.g ["time", "level", "message"]

    :type topic_format: string
    :param topic_format: "none" or "group_topic"

    :type filter_keys: string list
    :param filter_keys: the keys used to filter logs, e.g ["key_1", "key_2"]

    :type filter_keys_reg: string list
    :param filter_keys_reg: the regex for filter_keys to filter the log, filter_keys_reg[i] is for filter_keys[i]. The size of filter_keys_reg and filter_keys should be same. If a log is matched only if the size of filter_keys is 0, or all the value of the related keys in filter_keys, match the regex set in filter_keys_reg

    :type logSample: string
    :param logSample: sample strings for the log, up to 1000 bytes

    :type log_type: string
    :param log_type: common_reg_log, delimiter_log, apsara_log etc.

    :type extended_items: dict
    :param extended_items: extended items in dict format e.g. enableRawLog etc. refer to wiki page.
    https://help.aliyun.com/document_detail/29042.html?spm=5176.doc28997.6.744.C583Jg

    """

    def __init__(self, config_name, logstore_name, endpoint, log_path, file_pattern, time_format, log_begin_regex,
                 log_parse_regex, reg_keys,
                 topic_format="none", filter_keys=None, filter_keys_reg=None, logSample='',
                 log_type='common_reg_log', **extended_items):
        logger.warning("aliyun.log.CommonRegLogConfigDetail is deprecated and will be removed in future version."
                      "Use ConfigDetailBase based class instead")

        if filter_keys is None:
            filter_keys = []
        if filter_keys_reg is None:
            filter_keys_reg = []

        LogtailConfigDetail.__init__(self, config_name, logstore_name, endpoint, log_path, file_pattern,
                                     log_begin_regex,
                                     topic_format, filter_keys, filter_keys_reg, logSample, log_type, **extended_items)

        self.time_format = time_format
        self.log_parse_regex = log_parse_regex
        self.keys = reg_keys

    def to_json(self):
        json_value = {"configName": self.config_name, "inputType": "file"}

        # add log sample
        if self.logSample:
            json_value["logSample"] = self.logSample

        detail = {'logType': self.log_type, 'logPath': self.log_path, 'filePattern': self.file_pattern,
                  'localStorage': True, 'timeFormat': self.time_format, 'logBeginRegex': self.log_begin_regex,
                  'regex': self.log_parse_regex, 'key': self.keys, 'filterKey': self.filter_keys,
                  'filterRegex': self.filter_keys_reg, 'topicFormat': self.topic_format}
        detail.update(self.extended_items)
        json_value["inputDetail"] = detail
        json_value["outputType"] = "LogService"
        output_detail = {}
        if self.endpoint:
            output_detail["endpoint"] = self.endpoint
        output_detail["logstoreName"] = self.logstore_name
        json_value['outputDetail'] = output_detail
        return json_value


class ApsaraLogConfigDetail(LogtailConfigDetail):
    """The logtail config for apsara_log
    :type config_name: string
    :param config_name: the config name

    :type logstore_name: string
    :param logstore_name: the logstore name for the config 

    :type endpoint: string
    :param endpoint: deprecated, set it as empty

    :type log_path: string
    :param log_path: the log file dir path

    :type file_pattern: string
    :param file_pattern: the log file name pattern, e.g *.LOG , access.log

    :type log_begin_regex: string
    :param log_begin_regex: the regular express to match the first line of a log

    :type topic_format: string
    :param topic_format: "none" or "group_topic"

    :type filter_keys: string list
    :param filter_keys: the keys used to filter logs, e.g ["key_1", "key_2"]

    :type filter_keys_reg: string list
    :param filter_keys_reg: the regex for filter_keys to filter the log, filter_keys_reg[i] is for filter_keys[i]. The size of filter_keys_reg and filter_keys should be same. If a log is matched only if the size of filter_keys is 0, or all the value of the related keys in filter_keys, match the regex set in filter_keys_reg

    :type extended_items: dict
    :param extended_items: extended items in dict format e.g. enableRawLog etc. refer to wiki page.
    https://help.aliyun.com/document_detail/29042.html?spm=5176.doc28997.6.744.C583Jg

    """

    def __init__(self, config_name, logstore_name, endpoint, log_path, file_pattern,
                 log_begin_regex=r'\[\d+-\d+-\d+ \d+:\d+:\d+\.\d+.*', topic_format="none", filter_keys=None,
                 filter_keys_reg=None, logSample='', **extended_items):
        logger.warning("aliyun.log.ApsaraLogConfigDetail is deprecated and will be removed in future version. "
                      "Use ConfigDetailBase based class instead")

        if filter_keys_reg is None:
            filter_keys_reg = []
        if filter_keys is None:
            filter_keys = []
        LogtailConfigDetail.__init__(self, config_name, logstore_name, endpoint, log_path, file_pattern,
                                     log_begin_regex,
                                     topic_format, filter_keys, filter_keys_reg, logSample,
                                     'apsara_log', **extended_items)

    def to_json(self):
        json_value = {"configName": self.config_name, "inputType": "file"}
        # add log sample
        if self.logSample:
            json_value["logSample"] = self.logSample

        detail = {'logType': 'apsara_log', 'logPath': self.log_path, 'filePattern': self.file_pattern,
                  'localStorage': True, 'logBeginRegex': self.log_begin_regex, 'timeFormat': '',
                  'filterKey': self.filter_keys, 'filterRegex': self.filter_keys_reg, 'topicFormat': self.topic_format}
        detail.update(self.extended_items)  # add more extended items
        json_value["inputDetail"] = detail
        json_value["outputType"] = "log"
        output_detail = {}
        if self.endpoint is not None:
            output_detail["endpoint"] = self.endpoint
        output_detail["logstoreName"] = self.logstore_name
        json_value['outputDetail'] = output_detail
        return json_value


class LogtailConfigHelper(object):
    """
    A helper to generate logtail config object from dict object (loaded from json)
    """

    @staticmethod
    def generate_common_reg_log_config(json_value):
        """Generate common logtail config from loaded json value

        :param json_value:
        :return:
        """
        input_detail = copy.deepcopy(json_value['inputDetail'])
        output_detail = json_value['outputDetail']
        logSample = json_value.get('logSample', '')
        config_name = json_value['configName']
        logstore_name = output_detail['logstoreName']
        endpoint = output_detail.get('endpoint', '')

        log_path = input_detail['logPath']
        file_pattern = input_detail['filePattern']
        time_format = input_detail['timeFormat']
        log_begin_regex = input_detail.get('logBeginRegex', '')
        log_parse_regex = input_detail.get('regex', '')
        reg_keys = input_detail['key']
        topic_format = input_detail['topicFormat']
        filter_keys = input_detail['filterKey']
        filter_keys_reg = input_detail['filterRegex']
        log_type = input_detail.get('logType')

        for item in ('logPath', 'filePattern', 'timeFormat', 'logBeginRegex', 'regex', 'key',
                     'topicFormat', 'filterKey', 'filterRegex', 'logType'):
            if item in input_detail:
                del input_detail[item]

        config = CommonRegLogConfigDetail(config_name, logstore_name, endpoint, log_path, file_pattern, time_format,
                                          log_begin_regex, log_parse_regex, reg_keys,
                                          topic_format, filter_keys, filter_keys_reg, logSample,
                                          log_type, **input_detail)
        return config

    @staticmethod
    def generate_apsara_log_config(json_value):
        """Generate apsara logtail config from loaded json value

        :param json_value:
        :return:
        """
        input_detail = json_value['inputDetail']
        output_detail = json_value['outputDetail']
        config_name = json_value['configName']
        logSample = json_value.get('logSample', '')

        logstore_name = output_detail['logstoreName']
        endpoint = output_detail.get('endpoint', '')
        log_path = input_detail['logPath']
        file_pattern = input_detail['filePattern']

        log_begin_regex = input_detail.get('logBeginRegex', '')
        topic_format = input_detail['topicFormat']
        filter_keys = input_detail['filterKey']
        filter_keys_reg = input_detail['filterRegex']

        config = ApsaraLogConfigDetail(config_name, logstore_name, endpoint, log_path, file_pattern,
                                       log_begin_regex, topic_format, filter_keys, filter_keys_reg, logSample)
        return config

    @staticmethod
    def generate_logtail_config(json_value):
        """Generate logtail config from loaded json value

        :param json_value:
        :return:
        """
        logger.warning("aliyun.log.LogtailConfigHelper is deprecated and will be removed in future version."
                      "Use LogtailConfigGenerator instead")

        if json_value['inputDetail']['logType'] == 'apsara_log':
            return LogtailConfigHelper.generate_apsara_log_config(json_value)
        return LogtailConfigHelper.generate_common_reg_log_config(json_value)


class ConfigDetailBase(object):
    MANDATORY_FIELDS_ROOT = ["configName", "inputType", "inputDetail"]
    MANDATORY_FIELDS_DETAIL = []
    DEFAULT_DETAIL_FIELDS = [('localStorage', True)]

    def __init__(self, logstoreName, configName, inputType, createTime=None, modifyTime=None, logSample=None,
                 **input_detail):
        self.value = {
            "configName": configName,
            "logSample": logSample,
            "createTime": createTime,
            "modifyTime": modifyTime,
            "inputDetail": input_detail,
            "inputType": inputType,
            "outputDetail": {
                "logstoreName": logstoreName
            },
            "outputType": "LogService"
        }

        # clean up none items
        self.__clean_up_non_items()

        # add default ones
        for k, v in self.DEFAULT_DETAIL_FIELDS:
            if k not in self.value["inputDetail"]:
                self.value["inputDetail"][k] = v

    @property
    def config_name(self):
        return self.value["configName"]

    @config_name.setter
    def config_name(self, value):
        self.value["configName"] = value

    @property
    def logstore_name(self):
        return self.value["outputDetail"]["logstoreName"]

    @logstore_name.setter
    def logstore_name(self, value):
        self.value["outputDetail"]["logstoreName"] = value

    def __clean_up_non_items(self):
        none_items = [k for k, v in self.value.items() if v is None]
        for k in none_items:
            del self.value[k]
        none_detail_items = [k for k, v in self.value["inputDetail"].items() if v is None]
        for k in none_detail_items:
            del self.value["inputDetail"][k]

    @classmethod
    def from_json(cls, json_value):
        for item in cls.MANDATORY_FIELDS_ROOT:
            if item not in json_value:
                raise ValueError('item "{0}" not in json value'.format(item))

        for item in cls.MANDATORY_FIELDS_DETAIL:
            if item not in json_value["inputDetail"]:
                raise ValueError('item "{0}" not in json value "inputDetail"'.format(item))

        logstore_name = json_value["outputDetail"]["logstoreName"]
        config_name = json_value["configName"]
        input_type = json_value["inputType"]
        create_time = json_value.get("createTime", None)
        modify_time = json_value.get("modifyTime", None)
        log_sample = json_value.get("logSample", None)
        input_detail = json_value["inputDetail"]

        if cls == ConfigDetailBase:
            return cls(logstoreName=logstore_name, configName=config_name,
                       input_type=input_type,
                       createTime=create_time, modifyTime=modify_time,
                       logSample=log_sample, **input_detail)

        return cls(logstoreName=logstore_name, configName=config_name,
                   createTime=create_time, modifyTime=modify_time,
                   logSample=log_sample, **input_detail)

    def to_json(self):
        return self.value


class PluginConfigDetail(ConfigDetailBase):
    """The logtail config for simple mode

    :type logstoreName: string
    :param logstoreName: the logstore name

    :type configName: string
    :param configName: the config name

    :type logPath: string
    :param logPath: folder of log path /apsara/nuwa/

    :type filePattern: string
    :param filePattern: file path, e.g. *.log, it will be /apsara/nuwa/.../*.log

    :type localStorage: bool
    :param localStorage: if use local cache 1GB when logtail is offline. default is True.

    :type enableRawLog: bool
    :param enableRawLog: if upload raw data in content, default is False

    :type topicFormat: string
    :param topicFormat: "none", "group_topic" or regex to extract value from file path e.g. "/test/(\w+).log" will extract each file as topic, default is "none"

    :type fileEncoding: string
    :param fileEncoding: "utf8" or "gbk" so far

    :type maxDepth: int
    :param maxDepth: max depth of folder to scan, by default its 100, 0 means just scan the root folder

    :type preserve: bool
    :param preserve: if preserve time-out, by default is False, 30 min time-out if set it as True

    :type preserveDepth: int
    :param preserveDepth: time-out folder depth. 1-3

    :type filterKey: string list
    :param filterKey: only keep log which match the keys. e.g. ["city", "location"] will only scan files math the two fields

    :type filterRegex: string list
    :param filterRegex: matched value for filterKey, e.g. ["shanghai|beijing|nanjing", "east"] note, it's regex value list

    :type createTime: int
    :param createTime: timestamp of created, only useful when getting data from REST

    :type modifyTime: int
    :param modifyTime: timestamp of last modified time, only useful when getting data from REST

    :type extended_items: dict
    :param extended_items: extended items

    """
    MANDATORY_FIELDS_DETAIL = ConfigDetailBase.MANDATORY_FIELDS_DETAIL + ['plugin']
    DEFAULT_DETAIL_FIELDS = ConfigDetailBase.DEFAULT_DETAIL_FIELDS

    def __init__(self, logstoreName, configName, plugin, **extended_items):
        input_detail = {
            "plugin": plugin
        }
        input_detail.update(extended_items)

        ConfigDetailBase.__init__(self, logstoreName, configName, "plugin", **input_detail)


class SeperatorFileConfigDetail(ConfigDetailBase):
    """The logtail config for separator mode

    :type logstoreName: string
    :param logstoreName: the logstore name

    :type configName: string
    :param configName: the config name

    :type logPath: string
    :param logPath: folder of log path /apsara/nuwa/

    :type filePattern: string
    :param filePattern: file path, e.g. *.log, it will be /apsara/nuwa/.../*.log

    :type logSample: string
    :param logSample: log sample. e.g. shanghai|2000|east

    :type separator: string
    :param separator: '\t' for tab, ' ' for space, '|',  up to 3 chars like "&&&" or "||" etc.

    :type key: string list
    :param key: keys to map the fields like ["city", "population", "location"]

    :type timeKey: string
    :param timeKey: one key name in `key` to set the time or set it None to use system time.

    :type timeFormat: string
    :param timeFormat: whe timeKey is not None, set its format, refer to https://help.aliyun.com/document_detail/28980.html?spm=5176.2020520112.113.4.2243b18eHkxdNB

    :type localStorage: bool
    :param localStorage: if use local cache 1GB when logtail is offline. default is True.

    :type enableRawLog: bool
    :param enableRawLog: if upload raw data in content, default is False

    :type topicFormat: string
    :param topicFormat: "none", "group_topic" or regex to extract value from file path e.g. "/test/(\w+).log" will extract each file as topic, default is "none"

    :type fileEncoding: string
    :param fileEncoding: "utf8" or "gbk" so far

    :type maxDepth: int
    :param maxDepth: max depth of folder to scan, by default its 100, 0 means just scan the root folder

    :type preserve: bool
    :param preserve: if preserve time-out, by default is False, 30 min time-out if set it as True

    :type preserveDepth: int
    :param preserveDepth: time-out folder depth. 1-3

    :type filterKey: string list
    :param filterKey: only keep log which match the keys. e.g. ["city", "location"] will only scan files math the two fields

    :type filterRegex: string list
    :param filterRegex: matched value for filterKey, e.g. ["shanghai|beijing|nanjing", "east"] note, it's regex value list

    :type createTime: int
    :param createTime: timestamp of created, only useful when getting data from REST

    :type modifyTime: int
    :param modifyTime: timestamp of last modified time, only useful when getting data from REST

    :type extended_items: dict
    :param extended_items: extended items

    """
    MANDATORY_FIELDS_ROOT = ConfigDetailBase.MANDATORY_FIELDS_ROOT + ["logSample"]
    MANDATORY_FIELDS_DETAIL = ConfigDetailBase.MANDATORY_FIELDS_DETAIL \
                              + ["logPath", "filePattern", "separator", "key"]
    DEFAULT_DETAIL_FIELDS = [("logType", "delimiter_log"), ("localStorage", True),
                             ("timeFormat", ''), ("topicFormat", "none"), ("autoExtend", True)]

    def __init__(self, logstoreName, configName, logPath, filePattern, logSample, separator, key,
                 timeKey='', timeFormat=None, localStorage=None, enableRawLog=None, topicFormat=None,
                 fileEncoding=None, maxDepth=None, preserve=None, preserveDepth=None, filterKey=None,
                 filterRegex=None, createTime=None, modifyTime=None, **extended_items):
        input_detail = {
            "logPath": logPath,
            "filePattern": filePattern,
            "separator": separator,
            "key": key,
            "timeFormat": timeFormat,
            "timeKey": timeKey,
            "localStorage": localStorage,
            "enableRawLog": enableRawLog,
            "topicFormat": topicFormat,
            "fileEncoding": fileEncoding,
            "maxDepth": maxDepth,
            "preserve": preserve,
            "preserveDepth": preserveDepth,
            "filterKey": filterKey,
            "filterRegex": filterRegex
        }
        input_detail.update(extended_items)

        ConfigDetailBase.__init__(self, logstoreName, configName, "file", createTime=createTime, modifyTime=modifyTime,
                                  logSample=logSample, **input_detail)


class SimpleFileConfigDetail(ConfigDetailBase):
    """The logtail config for simple mode

    :type logstoreName: string
    :param logstoreName: the logstore name

    :type configName: string
    :param configName: the config name

    :type logPath: string
    :param logPath: folder of log path /apsara/nuwa/

    :type filePattern: string
    :param filePattern: file path, e.g. *.log, it will be /apsara/nuwa/.../*.log

    :type localStorage: bool
    :param localStorage: if use local cache 1GB when logtail is offline. default is True.

    :type enableRawLog: bool
    :param enableRawLog: if upload raw data in content, default is False

    :type topicFormat: string
    :param topicFormat: "none", "group_topic" or regex to extract value from file path e.g. "/test/(\w+).log" will extract each file as topic, default is "none"

    :type fileEncoding: string
    :param fileEncoding: "utf8" or "gbk" so far

    :type maxDepth: int
    :param maxDepth: max depth of folder to scan, by default its 100, 0 means just scan the root folder

    :type preserve: bool
    :param preserve: if preserve time-out, by default is False, 30 min time-out if set it as True

    :type preserveDepth: int
    :param preserveDepth: time-out folder depth. 1-3

    :type filterKey: string list
    :param filterKey: only keep log which match the keys. e.g. ["city", "location"] will only scan files math the two fields

    :type filterRegex: string list
    :param filterRegex: matched value for filterKey, e.g. ["shanghai|beijing|nanjing", "east"] note, it's regex value list

    :type createTime: int
    :param createTime: timestamp of created, only useful when getting data from REST

    :type modifyTime: int
    :param modifyTime: timestamp of last modified time, only useful when getting data from REST

    :type extended_items: dict
    :param extended_items: extended items

    """

    MANDATORY_FIELDS_DETAIL = ConfigDetailBase.MANDATORY_FIELDS_DETAIL \
                              + ["logPath", "filePattern"]
    DEFAULT_DETAIL_FIELDS = ConfigDetailBase.DEFAULT_DETAIL_FIELDS \
                            + [("logType", 'common_reg_log'), ("regex", '(.*)'), ("key", ['content']),
                               ("timeFormat", ''), ("topicFormat", "none")]

    def __init__(self, logstoreName, configName, logPath, filePattern, localStorage=None,
                 enableRawLog=None, topicFormat=None,
                 fileEncoding=None, maxDepth=None, preserve=None, preserveDepth=None, filterKey=None,
                 filterRegex=None, **extended_items):
        input_detail = {
            "logPath": logPath,
            "filePattern": filePattern,
            "localStorage": localStorage,
            "enableRawLog": enableRawLog,
            "topicFormat": topicFormat,
            "fileEncoding": fileEncoding,
            "maxDepth": maxDepth,
            "preserve": preserve,
            "preserveDepth": preserveDepth,
            "filterKey": filterKey,
            "filterRegex": filterRegex
        }
        input_detail.update(extended_items)

        ConfigDetailBase.__init__(self, logstoreName, configName, "file", **input_detail)


class FullRegFileConfigDetail(ConfigDetailBase):
    """The logtail config for full regex mode

    :type logstoreName: string
    :param logstoreName: the logstore name

    :type configName: string
    :param configName: the config name

    :type logPath: string
    :param logPath: folder of log path /apsara/nuwa/

    :type filePattern: string
    :param filePattern: file path, e.g. *.log, it will be /apsara/nuwa/.../*.log

    :type logSample: string
    :param logSample: log sample. e.g. shanghai|2000|east

    :type logBeginRegex: string
    :param logBeginRegex: regex to match line, None means '.*', just single line mode.

    :type regex: string
    :param regex: regex to extract fields form log. None means (.*), just capture whole line

    :type key: string list
    :param key: keys to map the fields like ["city", "population", "location"]. None means ["content"]

    :type timeFormat: string
    :param timeFormat: whe timeKey is not None, set its format, refer to https://help.aliyun.com/document_detail/28980.html?spm=5176.2020520112.113.4.2243b18eHkxdNB

    :type localStorage: bool
    :param localStorage: if use local cache 1GB when logtail is offline. default is True.

    :type enableRawLog: bool
    :param enableRawLog: if upload raw data in content, default is False

    :type topicFormat: string
    :param topicFormat: "none", "group_topic" or regex to extract value from file path e.g. "/test/(\w+).log" will extract each file as topic, default is "none"

    :type fileEncoding: string
    :param fileEncoding: "utf8" or "gbk" so far

    :type maxDepth: int
    :param maxDepth: max depth of folder to scan, by default its 100, 0 means just scan the root folder

    :type preserve: bool
    :param preserve: if preserve time-out, by default is False, 30 min time-out if set it as True

    :type preserveDepth: int
    :param preserveDepth: time-out folder depth. 1-3

    :type filterKey: string list
    :param filterKey: only keep log which match the keys. e.g. ["city", "location"] will only scan files math the two fields

    :type filterRegex: string list
    :param filterRegex: matched value for filterKey, e.g. ["shanghai|beijing|nanjing", "east"] note, it's regex value list

    :type createTime: int
    :param createTime: timestamp of created, only useful when getting data from REST

    :type modifyTime: int
    :param modifyTime: timestamp of last modified time, only useful when getting data from REST

    :type extended_items: dict
    :param extended_items: extended items

    """
    MANDATORY_FIELDS_DETAIL = ConfigDetailBase.MANDATORY_FIELDS_DETAIL \
                              + ["logPath", "filePattern"]
    DEFAULT_DETAIL_FIELDS = ConfigDetailBase.DEFAULT_DETAIL_FIELDS \
                            + [("logType", 'common_reg_log'), ("regex", '(.*)'),
                               ("key", ['content']), ("timeFormat", '')]

    def __init__(self, logstoreName, configName, logPath, filePattern, logSample,
                 logBeginRegex=None, regex=None, key=None, timeFormat=None,
                 localStorage=None,
                 enableRawLog=None, topicFormat=None,
                 fileEncoding=None, maxDepth=None, preserve=None, preserveDepth=None, filterKey=None,
                 filterRegex=None, **extended_items):
        input_detail = {
            "logPath": logPath,
            "filePattern": filePattern,
            "logBeginRegex": logBeginRegex,
            "regex": regex,
            "key": key,
            "timeFormat": timeFormat,
            "localStorage": localStorage,
            "enableRawLog": enableRawLog,
            "topicFormat": topicFormat,
            "fileEncoding": fileEncoding,
            "maxDepth": maxDepth,
            "preserve": preserve,
            "preserveDepth": preserveDepth,
            "filterKey": filterKey,
            "filterRegex": filterRegex
        }
        input_detail.update(extended_items)

        ConfigDetailBase.__init__(self, logstoreName, configName, "file", logSample=logSample, **input_detail)


class JsonFileConfigDetail(ConfigDetailBase):
    """The logtail config for json mode

    :type logstoreName: string
    :param logstoreName: the logstore name

    :type configName: string
    :param configName: the config name

    :type logPath: string
    :param logPath: folder of log path /apsara/nuwa/

    :type filePattern: string
    :param filePattern: file path, e.g. *.log, it will be /apsara/nuwa/.../*.log

    :type timeKey: string
    :param timeKey: one key name in `key` to set the time or set it None to use system time.

    :type timeFormat: string
    :param timeFormat: whe timeKey is not None, set its format, refer to https://help.aliyun.com/document_detail/28980.html?spm=5176.2020520112.113.4.2243b18eHkxdNB

    :type localStorage: bool
    :param localStorage: if use local cache 1GB when logtail is offline. default is True.

    :type enableRawLog: bool
    :param enableRawLog: if upload raw data in content, default is False

    :type topicFormat: string
    :param topicFormat: "none", "group_topic" or regex to extract value from file path e.g. "/test/(\w+).log" will extract each file as topic, default is "none"

    :type fileEncoding: string
    :param fileEncoding: "utf8" or "gbk" so far

    :type maxDepth: int
    :param maxDepth: max depth of folder to scan, by default its 100, 0 means just scan the root folder

    :type preserve: bool
    :param preserve: if preserve time-out, by default is False, 30 min time-out if set it as True

    :type preserveDepth: int
    :param preserveDepth: time-out folder depth. 1-3

    :type filterKey: string list
    :param filterKey: only keep log which match the keys. e.g. ["city", "location"] will only scan files math the two fields

    :type filterRegex: string list
    :param filterRegex: matched value for filterKey, e.g. ["shanghai|beijing|nanjing", "east"] note, it's regex value list

    :type createTime: int
    :param createTime: timestamp of created, only useful when getting data from REST

    :type modifyTime: int
    :param modifyTime: timestamp of last modified time, only useful when getting data from REST

    :type extended_items: dict
    :param extended_items: extended items

    """
    MANDATORY_FIELDS_DETAIL = ConfigDetailBase.MANDATORY_FIELDS_DETAIL \
                              + ["logPath", "filePattern"]
    DEFAULT_DETAIL_FIELDS = ConfigDetailBase.DEFAULT_DETAIL_FIELDS \
                            + [("logType", "json_log"), ('localStorage', True),
                               ("timeFormat", ''), ("topicFormat", "none")]

    def __init__(self, logstoreName, configName, logPath, filePattern, timeKey='', timeFormat=None,
                 localStorage=None, enableRawLog=None, topicFormat=None,
                 fileEncoding=None, maxDepth=None, preserve=None, preserveDepth=None, filterKey=None,
                 filterRegex=None, createTime=None, modifyTime=None, **extended_items):
        input_detail = {
            "logPath": logPath,
            "filePattern": filePattern,
            "timeFormat": timeFormat,
            "timeKey": timeKey,
            "localStorage": localStorage,
            "enableRawLog": enableRawLog,
            "topicFormat": topicFormat,
            "fileEncoding": fileEncoding,
            "maxDepth": maxDepth,
            "preserve": preserve,
            "preserveDepth": preserveDepth,
            "filterKey": filterKey,
            "filterRegex": filterRegex
        }
        input_detail.update(extended_items)

        ConfigDetailBase.__init__(self, logstoreName, configName, "file", createTime=createTime, modifyTime=modifyTime,
                                  **input_detail)


class ApsaraFileConfigDetail(ConfigDetailBase):
    """The logtail config for Apsara mode

    :type logstoreName: string
    :param logstoreName: the logstore name

    :type configName: string
    :param configName: the config name

    :type logPath: string
    :param logPath: folder of log path /apsara/nuwa/

    :type filePattern: string
    :param filePattern: file path, e.g. *.log, it will be /apsara/nuwa/.../*.log

    :type logBeginRegex: string
    :param logBeginRegex: regex to match line, None means '.*', just single line mode.

    :type localStorage: bool
    :param localStorage: if use local cache 1GB when logtail is offline. default is True.

    :type enableRawLog: bool
    :param enableRawLog: if upload raw data in content, default is False

    :type topicFormat: string
    :param topicFormat: "none", "group_topic" or regex to extract value from file path e.g. "/test/(\w+).log" will extract each file as topic, default is "none"

    :type fileEncoding: string
    :param fileEncoding: "utf8" or "gbk" so far

    :type maxDepth: int
    :param maxDepth: max depth of folder to scan, by default its 100, 0 means just scan the root folder

    :type preserve: bool
    :param preserve: if preserve time-out, by default is False, 30 min time-out if set it as True

    :type preserveDepth: int
    :param preserveDepth: time-out folder depth. 1-3

    :type filterKey: string list
    :param filterKey: only keep log which match the keys. e.g. ["city", "location"] will only scan files math the two fields

    :type filterRegex: string list
    :param filterRegex: matched value for filterKey, e.g. ["shanghai|beijing|nanjing", "east"] note, it's regex value list

    :type createTime: int
    :param createTime: timestamp of created, only useful when getting data from REST

    :type modifyTime: int
    :param modifyTime: timestamp of last modified time, only useful when getting data from REST

    :type extended_items: dict
    :param extended_items: extended items

    """
    MANDATORY_FIELDS_DETAIL = ConfigDetailBase.MANDATORY_FIELDS_DETAIL \
                              + ["logPath", "filePattern", "logBeginRegex"]
    DEFAULT_DETAIL_FIELDS = ConfigDetailBase.DEFAULT_DETAIL_FIELDS \
                            + [("logType", "apsara_log"), ("topicFormat", "none")]

    def __init__(self, logstoreName, configName, logPath, filePattern, logBeginRegex,
                 localStorage=None, enableRawLog=None, topicFormat=None,
                 fileEncoding=None, maxDepth=None, preserve=None, preserveDepth=None, filterKey=None,
                 filterRegex=None, createTime=None, modifyTime=None, **extended_items):
        input_detail = {
            "logPath": logPath,
            "filePattern": filePattern,
            "logBeginRegex": logBeginRegex,
            "localStorage": localStorage,
            "enableRawLog": enableRawLog,
            "topicFormat": topicFormat,
            "fileEncoding": fileEncoding,
            "maxDepth": maxDepth,
            "preserve": preserve,
            "preserveDepth": preserveDepth,
            "filterKey": filterKey,
            "filterRegex": filterRegex
        }
        input_detail.update(extended_items)

        ConfigDetailBase.__init__(self, logstoreName, configName, "file", createTime=createTime, modifyTime=modifyTime,
                                  **input_detail)


class SyslogConfigDetail(ConfigDetailBase):
    """The logtail config for syslog mode

    :type logstoreName: string
    :param logstoreName: the logstore name

    :type configName: string
    :param configName: the config name

    :type tag: string
    :param tag: tag for the log captured

    :type localStorage: bool
    :param localStorage: if use local cache 1GB when logtail is offline. default is True.

    :type createTime: int
    :param createTime: timestamp of created, only useful when getting data from REST

    :type modifyTime: int
    :param modifyTime: timestamp of last modified time, only useful when getting data from REST

    :type extended_items: dict
    :param extended_items: extended items

    """
    MANDATORY_FIELDS_DETAIL = ConfigDetailBase.MANDATORY_FIELDS_DETAIL + ["tag"]

    def __init__(self, logstoreName, configName, tag,
                 localStorage=None, createTime=None, modifyTime=None, **extended_items):
        input_detail = {
            "tag": tag,
            "localStorage": localStorage
        }
        input_detail.update(extended_items)

        ConfigDetailBase.__init__(self, logstoreName, configName, "syslog", createTime=createTime,
                                  modifyTime=modifyTime,
                                  **input_detail)


class LogtailConfigGenerator(object):
    """
    Generator of Logtial config
    """

    @staticmethod
    def generate_simple_log_config(json_value):
        return SimpleFileConfigDetail.from_json(json_value)

    @staticmethod
    def generate_json_config(json_value):
        return JsonFileConfigDetail.from_json(json_value)

    @staticmethod
    def generate_syslog_config(json_value):
        return SyslogConfigDetail.from_json(json_value)

    @staticmethod
    def generate_separator_config(json_value):
        return SeperatorFileConfigDetail.from_json(json_value)

    @staticmethod
    def generate_full_regex_config(json_value):
        return FullRegFileConfigDetail.from_json(json_value)

    @staticmethod
    def generate_apsara_config(json_value):
        return ApsaraFileConfigDetail.from_json(json_value)

    @staticmethod
    def generate_plugin_config(json_value):
        return PluginConfigDetail.from_json(json_value)

    @staticmethod
    def generate_config(json_value):
        input_type = json_value.get("inputType", "")
        if input_type == LogtailInputType.PLUGIN.value:
            return LogtailConfigGenerator.generate_plugin_config(json_value)
        elif input_type == LogtailInputType.SYSLOG.value:
            return LogtailConfigGenerator.generate_syslog_config(json_value)
        elif input_type == LogtailInputType.FILE.value:
            log_type = json_value["inputDetail"].get("logType", "")
            if log_type == LogtailType.JSON.value:
                return LogtailConfigGenerator.generate_json_config(json_value)
            elif log_type == LogtailType.FULL_REGEX.value:
                # use simple which is also full actually
                return LogtailConfigGenerator.generate_simple_log_config(json_value)
            elif log_type == LogtailType.SEPARATOR.value:
                return LogtailConfigGenerator.generate_separator_config(json_value)
            elif log_type == LogtailType.APSARA.value:
                return LogtailConfigGenerator.generate_apsara_config(json_value)

        raise LogException("InvalidInput", "input json string missed necessary fields: "
                                           "input_type and/or logType")

    @staticmethod
    def from_json(json_value):
        return LogtailConfigGenerator.generate_config(json_value)




#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.

from .logrequest import LogRequest
from .util import parse_timestamp

class GetHistogramsRequest(LogRequest):
    """ The request used to get histograms of a query from log.
    
    :type project: string
    :param project: project name
    
    :type logstore: string
    :param logstore: logstore name
    
    :type fromTime: int/string
    :param fromTime: the begin time or format of time in readable time like "%Y-%m-%d %H:%M:%S<time_zone>" e.g. "2018-01-02 12:12:10+8:00" e.g. "2018-01-02 12:12:10", also support human readable string, e.g. "1 hour ago", "now", "yesterday 0:0:0", refer to https://aliyun-log-cli.readthedocs.io/en/latest/tutorials/tutorial_human_readable_datetime.html
    
    :type toTime: int/string
    :param toTime: the end time or format of time in readable time like "%Y-%m-%d %H:%M:%S<time_zone>" e.g. "2018-01-02 12:12:10+8:00" e.g. "2018-01-02 12:12:10", also support human readable string, e.g. "1 hour ago", "now", "yesterday 0:0:0", refer to https://aliyun-log-cli.readthedocs.io/en/latest/tutorials/tutorial_human_readable_datetime.html

    :type topic: string
    :param topic: topic name of logs
    
    :type query: string
    :param query: user defined query
    """

    def __init__(self, project=None, logstore=None, fromTime=None, toTime=None, topic=None, query=None):
        LogRequest.__init__(self, project)
        self.logstore = logstore
        self.fromTime = parse_timestamp(fromTime)
        self.toTime = parse_timestamp(toTime)
        self.topic = topic
        self.query = query

    def get_logstore(self):
        """ Get logstore name
        
        :return: string, logstore name.
        """
        return self.logstore if self.logstore else ''

    def set_logstore(self, logstore):
        """ Set logstore name
        
        :type logstore: string
        :param logstore: logstore name
        """
        self.logstore = logstore

    def get_topic(self):
        """ Get topic name
        
        :return: string, topic name
        """
        return self.topic if self.topic else ''

    def set_topic(self, topic):
        """ Set topic name
        
        :type topic: string
        :param topic: topic name
        """
        self.topic = topic

    def get_from(self):
        """ Get begin time
        
        :return: int, begin time
        """
        return self.fromTime

    def set_from(self, fromTime):
        """ Set begin time
        
        :type fromTime: int
        :param fromTime: begin time
        """
        self.fromTime = fromTime

    def get_to(self):
        """ Get end time
        
        :return: int, end time
        """
        return self.toTime

    def set_to(self, toTime):
        """ Set end time
        
        :type toTime: int
        :param toTime: end time
        """
        self.toTime = toTime

    def get_query(self):
        """ Get user defined query
        
        :return: string, user defined query
        """
        return self.query

    def set_query(self, query):
        """ Set user defined query
        
        :type query: string
        :param query: user defined query
        """
        self.query = query



from __future__ import print_function
from .logexception import LogException
import six
import json
import os
from concurrent.futures import ProcessPoolExecutor, as_completed
import multiprocessing
from .logresponse import LogResponse
from json import JSONEncoder
import logging
from collections import defaultdict
import time
from .etl_core import Runner
from .putlogsrequest import PutLogsRequest
from .logitem import LogItem
from aliyun.log.pulllog_response import PullLogResponse
from .consumer import *
from multiprocessing import RLock
from .util import base64_encodestring as b64e
import copy


MAX_INIT_SHARD_COUNT = 200

logger = logging.getLogger(__name__)


def copy_project(from_client, to_client, from_project, to_project, copy_machine_group=False):
    """
    copy project, logstore, machine group and logtail config to target project,
    will create the target project if it doesn't exist

    :type from_client: LogClient
    :param from_client: logclient instance

    :type to_client: LogClient
    :param to_client: logclient instance

    :type from_project: string
    :param from_project: project name

    :type to_project: string
    :param to_project: project name

    :type copy_machine_group: bool
    :param copy_machine_group: if copy machine group resources, False by default.

    :return:
    """

    # copy project
    ret = from_client.get_project(from_project)
    try:
        ret = to_client.create_project(to_project, ret.get_description())
    except LogException as ex:
        if ex.get_error_code() == 'ProjectAlreadyExist':
            # don't create the project as it already exists
            pass
        else:
            raise

    default_fetch_size = 100

    # list logstore and copy them
    offset, size = 0, default_fetch_size
    source_logstores, all_source_logstores = set(), set()
    while True:
        ret = from_client.list_logstore(from_project, offset=offset, size=size)
        count = ret.get_logstores_count()
        total = ret.get_logstores_total()
        source_logstores = set(ret.get_logstores())
        all_source_logstores = all_source_logstores.union(source_logstores)
        for logstore_name in source_logstores:
            # copy logstore
            ret = from_client.get_logstore(from_project, logstore_name)
            res_shard = from_client.list_shards(from_project, logstore_name)
            expected_rwshard_count = len([shard for shard in res_shard.shards if shard['status'].lower() == 'readwrite'])
            try:
                ret2 = to_client.create_logstore(to_project, logstore_name, ret.get_ttl(),
                                                min(expected_rwshard_count, MAX_INIT_SHARD_COUNT),
                                                enable_tracking=ret.get_enable_tracking(),
                                                append_meta=ret.append_meta,
                                                auto_split=ret.auto_split,
                                                max_split_shard=ret.max_split_shard,
                                                preserve_storage=ret.preserve_storage
                                                )
            except LogException as ex:
                if ex.get_error_code().lower() == "logstorealreadyexist":
                    pass
                else:
                    raise

            # copy index
            index_config = None
            try:
                ret = from_client.get_index_config(from_project, logstore_name)
                index_config = ret.get_index_config()
            except LogException as ex:
                if ex.get_error_code() == 'IndexConfigNotExist':
                    pass
                else:
                    raise

            if index_config is not None:
                for x in range(60):
                    try:
                        ret2 = to_client.create_index(to_project, logstore_name, ret.get_index_config())
                        break
                    except LogException as ex:
                        if ex.get_error_code().lower() == "logstorenotexist" and x < 59:
                            time.sleep(1)
                            continue
                        if ex.get_error_code().lower() == "indexalreadyexist":
                            # target already has index, overwrite it
                            ret2 = to_client.update_index(to_project, logstore_name, index_config)
                            break
                        raise ex

        offset += count
        if count < size or offset >= total:
            break

    # list logtail config and copy them
    offset, size = 0, default_fetch_size
    source_configs = set()
    while True:
        ret = from_client.list_logtail_config(from_project, offset=offset, size=size)
        count = ret.get_configs_count()
        total = ret.get_configs_total()

        for config_name in ret.get_configs():
            ret = from_client.get_logtail_config(from_project, config_name)
            if ret.logtail_config.logstore_name not in all_source_logstores:
                continue

            source_configs.add(config_name)
            for x in range(60):
                try:
                    ret2 = to_client.create_logtail_config(to_project, ret.logtail_config)
                except LogException as ex:
                    if ex.get_error_code().lower() == "logstorenotexist" and x < 59:
                        time.sleep(1)
                        continue
                    if ex.get_error_code().lower() == "configalreadyexist":
                        break
                    raise ex

        offset += count
        if count < size or offset >= total:
            break

    # list machine group and copy them
    offset, size = 0, default_fetch_size
    while copy_machine_group:
        ret = from_client.list_machine_group(from_project, offset=offset, size=size)
        count = ret.get_machine_group_count()
        total = ret.get_machine_group_total()

        for group_name in ret.get_machine_group():
            ret = from_client.get_machine_group(from_project, group_name)
            try:
                ret = to_client.create_machine_group(to_project, ret.get_machine_group())
            except LogException as ex:
                if ex.get_error_code() == 'MachineGroupAlreadyExist':
                    pass
                else:
                    raise ex

            # list all applied config and copy the relationship
            ret = from_client.get_machine_group_applied_configs(from_project, group_name)
            for config_name in ret.get_configs():
                if config_name not in source_configs:
                    continue

                for x in range(60):
                    try:
                       to_client.apply_config_to_machine_group(to_project, config_name, group_name)
                    except LogException as ex:
                        if ex.get_error_code().lower() in ("confignotexist", "groupnotexist") and x < 59:
                            time.sleep(1)
                            continue
                        raise ex

        offset += count
        if count < size or offset >= total:
            break


def copy_logstore(from_client, from_project, from_logstore, to_logstore, to_project=None, to_client=None, to_region_endpoint=None):
    """
    copy logstore, index, logtail config to target logstore, machine group are not included yet.
    the target logstore will be crated if not existing

    :type from_client: LogClient
    :param from_client: logclient instance

    :type from_project: string
    :param from_project: project name

    :type from_logstore: string
    :param from_logstore: logstore name

    :type to_logstore: string
    :param to_logstore: target logstore name

    :type to_project: string
    :param to_project: project name, copy to same project if not being specified, will try to create it if not being specified

    :type to_client: LogClient
    :param to_client: logclient instance, use it to operate on the "to_project" if being specified

    :type to_region_endpoint: string
    :param to_region_endpoint: target region, use it to operate on the "to_project" while "to_client" not be specified

    :return:
    """

    if to_region_endpoint is not None and to_client is None:
        to_client = copy.deepcopy(from_client)
        to_client.set_endpoint(to_region_endpoint)
    else:
        to_client = to_client or from_client

    # check client
    if to_project is not None:
        # check if target project exists or not
        ret = from_client.get_project(from_project)
        try:
            ret = to_client.create_project(to_project, ret.get_description())
        except LogException as ex:
            if ex.get_error_code() == 'ProjectAlreadyExist':
                # don't create the project as it already exists
                pass
            else:
                raise

    to_project = to_project or from_project

    # return if logstore are the same one
    if from_client is to_client and from_project == to_project and from_logstore == to_logstore:
        return

    # copy logstore
    ret = from_client.get_logstore(from_project, from_logstore)
    res_shard = from_client.list_shards(from_project, from_logstore)
    expected_rwshard_count = len([shard for shard in res_shard.shards if shard['status'].lower() == 'readwrite'])
    try:
        ret = to_client.create_logstore(to_project, to_logstore,
                                        ttl=ret.get_ttl(),
                                        shard_count=min(expected_rwshard_count, MAX_INIT_SHARD_COUNT),
                                        enable_tracking=ret.get_enable_tracking(),
                                        append_meta=ret.append_meta,
                                        auto_split=ret.auto_split,
                                        max_split_shard=ret.max_split_shard,
                                        preserve_storage=ret.preserve_storage)
    except LogException as ex:
        if ex.get_error_code().lower() == "logstorealreadyexist":
            # update logstore's settings
            ret = to_client.update_logstore(to_project, to_logstore,
                                            ttl=ret.get_ttl(),
                                            enable_tracking=ret.get_enable_tracking(),
                                            append_meta=ret.append_meta,
                                            auto_split=ret.auto_split,
                                            max_split_shard=ret.max_split_shard,
                                            preserve_storage=ret.preserve_storage
                                            )

            # arrange shard to expected count
            res = arrange_shard(to_client, to_project, to_logstore, min(expected_rwshard_count, MAX_INIT_SHARD_COUNT))
        else:
            raise


    # copy index
    try:
        ret = from_client.get_index_config(from_project, from_logstore)
        ret = to_client.create_index(to_project, to_logstore, ret.get_index_config())
    except LogException as ex:
        if ex.get_error_code() == 'IndexConfigNotExist':
            # source has no index
            pass
        elif ex.get_error_code() == 'IndexAlreadyExist':
            # target already has index, overwrite it
            ret = to_client.update_index(to_project, to_logstore, ret.get_index_config())
            pass
        else:
            raise

    # list logtail config linked to the logstore and copy them
    default_fetch_size = 100
    offset, size = 0, default_fetch_size
    while True:
        ret = from_client.list_logtail_config(from_project, offset=offset, size=size)
        count = ret.get_configs_count()
        total = ret.get_configs_total()

        for config_name in ret.get_configs():
            ret = from_client.get_logtail_config(from_project, config_name)
            config = ret.logtail_config
            if config.logstore_name != from_logstore:
                continue

            config.config_name = to_logstore + '_' + config_name
            config.logstore_name = to_logstore
            ret = to_client.create_logtail_config(to_project, config)

        offset += count
        if count < size or offset >= total:
            break


def list_more(fn, offset, size, batch_size, *args):
    """list all data using the fn
    """
    if size < 0:
        expected_total_size = six.MAXSIZE
    else:
        expected_total_size = size
        batch_size = min(size, batch_size)

    response = None
    total_count_got = 0
    while True:
        ret = fn(*args, offset=offset, size=batch_size)
        if response is None:
            response = ret
        else:
            response.merge(ret)

        count = ret.get_count()
        total = ret.get_total()
        offset += count
        total_count_got += count
        batch_size = min(batch_size, expected_total_size - total_count_got)

        if count == 0 or offset >= total or total_count_got >= expected_total_size:
            break

    return response


def query_more(fn, offset, size, batch_size, *args):
    """list all data using the fn
    """
    if size < 0:
        expected_total_size = six.MAXSIZE
    else:
        expected_total_size = size
        batch_size = min(size, batch_size)

    response = None
    total_count_got = 0
    complete = False
    while True:
        ret = fn(*args, offset=offset, size=batch_size)

        if response is None:
            response = ret
        else:
            response.merge(ret)

        # if incompete, exit
        if not ret.is_completed():
            break

        count = ret.get_count()
        offset += count
        total_count_got += count
        batch_size = min(batch_size, expected_total_size - total_count_got)
        if count == 0 or total_count_got >= expected_total_size:
            break

    return response


def get_encoder_cls(encodings):
    class NonUtf8Encoder(JSONEncoder):
        def default(self, obj):
            if isinstance(obj, six.binary_type):
                for encoding in encodings:
                    try:
                        return obj.decode(encoding)
                    except UnicodeDecodeError as ex:
                        pass
                return obj.decode('utf8', "ignore")

            return JSONEncoder.default(self, obj)

    return NonUtf8Encoder


def dump_worker(client, project_name, logstore_name, from_time, to_time,
                shard_id, file_path,
                batch_size=None, compress=None, encodings=None, no_escape=None):
    res = client.pull_log(project_name, logstore_name, shard_id, from_time, to_time, batch_size=batch_size,
                          compress=compress)
    encodings = encodings or ('utf8', 'latin1', 'gbk')
    ensure_ansi = not no_escape

    count = 0
    next_cursor = 'as from_time configured'
    try:
        for data in res:
            for log in data.get_flatten_logs_json(decode_bytes=True):
                with open(os.path.expanduser(file_path), "a+") as f:
                    count += 1
                    try:
                        if six.PY2:
                            last_ex = None
                            for encoding in encodings:
                                try:
                                    ret = json.dumps(log, encoding=encoding, ensure_ascii=ensure_ansi)
                                    if isinstance(ret, unicode):
                                        ret = ret.encode(encoding, errors="ignore")
                                    f.write(ret)
                                    f.write("\n")
                                    break
                                except UnicodeDecodeError as ex:
                                    last_ex = ex
                            else:
                                raise last_ex
                        else:
                            f.write(json.dumps(log, cls=get_encoder_cls(encodings), ensure_ascii=ensure_ansi))
                            f.write("\n")
                    except Exception as ex:
                        logger.error("shard: {0} Fail to dump log: {1}".format(shard_id, b64e(repr(log))), exc_info=True)
                        raise ex
            next_cursor = data.next_cursor
    except Exception as ex:
        logger.error("dump log failed: task info {0} failed to copy data to target, next cursor: {1} detail: {2}".
                     format(
            (project_name, logstore_name, shard_id, from_time, to_time),
            next_cursor, ex), exc_info=True)
        raise

    return file_path, count


def pull_log_dump(client, project_name, logstore_name, from_time, to_time, file_path, batch_size=None, compress=None,
                  encodings=None, shard_list=None, no_escape=None):
    cpu_count = multiprocessing.cpu_count() * 2

    shards = client.list_shards(project_name, logstore_name).get_shards_info()
    current_shards = [str(shard['shardID']) for shard in shards]
    target_shards = _parse_shard_list(shard_list, current_shards)
    worker_size = min(cpu_count, len(target_shards))

    result = dict()
    total_count = 0
    with ProcessPoolExecutor(max_workers=worker_size) as pool:
        futures = [pool.submit(dump_worker, client, project_name, logstore_name, from_time, to_time,
                               shard_id=shard, file_path=file_path.format(shard),
                               batch_size=batch_size, compress=compress, encodings=encodings, no_escape=no_escape)
                   for shard in target_shards]

        for future in as_completed(futures):
            file_path, count = future.result()
            total_count += count
            if count:
                result[file_path] = count

    return LogResponse({}, {"total_count": total_count, "files": result})


def copy_worker(from_client, from_project, from_logstore, shard_id, from_time, to_time,
                to_client, to_project, to_logstore, batch_size=None, compress=None,
                new_topic=None, new_source=None):
    next_cursor = "As from_time configured"

    try:
        iter_data = from_client.pull_log(from_project, from_logstore, shard_id, from_time, to_time, batch_size=batch_size,
                                         compress=compress)

        count = 0
        for res in iter_data:
            for loggroup in res.get_loggroup_list().LogGroups:
                if new_topic is not None:
                    loggroup.Topic = new_topic
                if new_source is not None:
                    loggroup.Source = new_source
                rtn = to_client.put_log_raw(to_project, to_logstore, loggroup, compress=compress)
                count += len(loggroup.Logs)

            next_cursor = res.next_cursor
        return shard_id, count
    except Exception as ex:
        logger.error("copy data failed: task info {0} failed to copy data to target, next cursor: {1} detail: {2}".
                     format( (from_project, from_logstore, shard_id, from_time, to_time, to_client, to_project, to_logstore), next_cursor, ex), exc_info=True)
        raise


def _parse_shard_list(shard_list, current_shard_list):
    """
    parse shard list
    :param shard_list: format like: 1,5-10,20
    :param current_shard_list: current shard list
    :return:
    """
    if not shard_list:
        return current_shard_list
    target_shards = []
    for n in shard_list.split(","):
        n = n.strip()
        if n.isdigit() and n in current_shard_list:
            target_shards.append(n)
        elif n:
            rng = n.split("-")
            if len(rng) == 2:
                s = rng[0].strip()
                e = rng[1].strip()
                if s.isdigit() and e.isdigit():
                    for x in range(int(s), int(e)+1):
                        if str(x) in current_shard_list:
                            target_shards.append(str(x))

    logger.info("parse_shard, shard_list: '{0}' current shard '{1}' result: '{2}'".format(shard_list,
                                                                                          current_shard_list, target_shards))
    if not target_shards:
        raise LogException("InvalidParameter", "There's no available shard with settings {0}".format(shard_list))

    return target_shards


def copy_data(from_client, from_project, from_logstore, from_time, to_time=None,
              to_client=None, to_project=None, to_logstore=None,
              shard_list=None,
              batch_size=None, compress=None, new_topic=None, new_source=None):
    """
    copy data from one logstore to another one (could be the same or in different region), the time is log received time on server side.

    """
    to_client = to_client or from_client
    # increase the timeout to 2 min at least
    from_client.timeout = max(from_client.timeout, 120)
    to_client.timeout = max(to_client.timeout, 120)

    to_project = to_project or from_project
    to_logstore = to_logstore or from_logstore
    to_time = to_time or "end"

    cpu_count = multiprocessing.cpu_count() * 2
    shards = from_client.list_shards(from_project, from_logstore).get_shards_info()
    current_shards = [str(shard['shardID']) for shard in shards]
    target_shards = _parse_shard_list(shard_list, current_shards)
    worker_size = min(cpu_count, len(target_shards))

    result = dict()
    total_count = 0
    with ProcessPoolExecutor(max_workers=worker_size) as pool:
        futures = [pool.submit(copy_worker, from_client, from_project, from_logstore, shard,
                               from_time, to_time,
                               to_client, to_project, to_logstore,
                               batch_size=batch_size, compress=compress,
                               new_topic=new_topic, new_source=new_source)
                   for shard in target_shards]

        for future in as_completed(futures):
            partition, count = future.result()
            total_count += count
            if count:
                result[partition] = count

    return LogResponse({}, {"total_count": total_count, "shards": result})


TOTAL_SHARD_SIZE = int('0xffffffffffffffffffffffffffffffff', base=16)
TOTAL_HASH_LENGTH = len('0xffffffffffffffffffffffffffffffff') - 2


def _split_one_shard_to_multiple(client, project, logstore, shard_info, count, current_shard_count):
    """return new_rw_shards_list, increased_shard_count """
    distance = shard_info['length'] // count
    if distance <= 0 or count <= 1:
        return [shard_info['info']], 0

    rw_shards, increased_shard_count = {shard_info['id']: shard_info['info']}, 0
    for x in range(1, count):
        new_hash = shard_info['start'] + distance * x
        new_hash = hex(new_hash)[2:].strip('lL')
        new_hash = '0' * (TOTAL_HASH_LENGTH - len(new_hash)) + new_hash
        try:
            if x == 1:
                res = client.split_shard(project, logstore, shard_info['id'], new_hash)
            else:
                res = client.split_shard(project, logstore, current_shard_count - 1, new_hash)

            # new rw_shards
            for shard in res.shards:
                if shard['status'] == 'readonly':
                    del rw_shards[shard['shardID']]
                else:
                    rw_shards[shard['shardID']] = shard

            current_shard_count += res.count - 1
            increased_shard_count += res.count - 1
            logger.info("split shard: project={0}, logstore={1}, shard_info={2}, count={3}, current_shard_count={4}".format(project, logstore, shard_info, count, current_shard_count))
        except Exception as ex:
            print(ex)
            print(x, project, logstore, shard_info, count, current_shard_count)
            raise

    return rw_shards.values(), increased_shard_count


def arrange_shard(client, project, logstore, count, stategy=None):
    total_length = TOTAL_SHARD_SIZE
    avg_len = total_length * 1.0 / count

    res = client.list_shards(project, logstore)
    current_shard_count = res.count
    rw_shards = [shard for shard in res.shards if shard['status'].lower() == 'readwrite']
    split_left = count - len(rw_shards)

    while split_left > 0:
        current_rw_shards = [{'id': shard['shardID'],
                              'start': int('0x' + shard['inclusiveBeginKey'], base=16),
                              'end': int('0x' + shard['exclusiveEndKey'], base=16),
                              'length': int('0x' + shard['exclusiveEndKey'], base=16)
                                        - int('0x' + shard['inclusiveBeginKey'], base=16),
                              'info': shard}
                             for shard in rw_shards]
        # need to split shard
        updated_rw_shards = []
        for i, shard in enumerate(sorted(current_rw_shards, key=lambda x: x['length'], reverse=True)):
            if split_left <= 0:  # no need to split any more
                break

            sp_cnt = int(shard['length'] // avg_len)
            if sp_cnt <= 1 and i == 0:  # cannot split, but be the first one, should split it
                sp_cnt = 2
            elif sp_cnt <= 0:
                sp_cnt = 1

            new_rw_shards, increased_shard_count = _split_one_shard_to_multiple(client, project, logstore, shard,
                                                                                sp_cnt, current_shard_count)
            updated_rw_shards.extend(new_rw_shards)
            current_shard_count += increased_shard_count
            split_left -= len(new_rw_shards) - 1

        # update current rw shards
        rw_shards = updated_rw_shards

    return ''


def _get_percentage(v, t, digit=2):
    return str(round(v * 100.0 / t, digit)) + '%'


class ResourceUsageResponse(LogResponse):
    def __init__(self, usage):
        LogResponse.__init__(self, {}, body=usage)


def get_resource_usage(client, project):
    result = {"logstore": {},
              "shard": {"logstores": {}},
              "logtail": {},
              "consumer_group": {"logstores": {}},
              }
    res = client.list_logstore(project, size=-1)
    result["logstore"] = {"count": {"status": res.total,
                                    "limitation": 200,
                                    "usage": _get_percentage(res.total, 200)}}
    shard_count = 0
    consumer_group_count = 0
    for logstore in res.logstores:
        res = client.list_shards(project, logstore)
        shard_count += res.count
        result["shard"]["logstores"][logstore] = {"status": res.count}

        res = client.list_consumer_group(project, logstore)
        if res.count:
            result["consumer_group"]["logstores"][logstore] = {"status": res.count, "limitation": 10,
                                                               "usage": _get_percentage(res.count, 10)}
        consumer_group_count += res.count

    result["shard"]["count"] = {"status": shard_count, "limitation": 200, "usage": _get_percentage(shard_count, 200)}
    result["consumer_group"]["count"] = {"status": consumer_group_count}

    res = client.list_logtail_config(project, offset=1000)  # pass 1000 to just get total count
    result["logtail"] = {"count": {"status": res.total,
                                   "limitation": 100,
                                   "usage": _get_percentage(res.total, 100)}}
    res = client.list_machine_group(project, offset=1000)  # pass 1000 to just get total count
    result["machine_group"] = {"count": {"status": res.total,
                                         "limitation": 100,
                                         "usage": _get_percentage(res.total, 100)}}
    res = client.list_savedsearch(project, offset=1000)  # pass 1000 to just get total count
    result["saved_search"] = {"count": {"status": res.total,
                                        "limitation": 100,
                                        "usage": _get_percentage(res.total, 100)}}

    res = client.list_dashboard(project, offset=1000)  # pass 1000 to just get total count
    result["dashboard"] = {"count": {"status": res.total,
                                     "limitation": 50,
                                     "usage": _get_percentage(res.total, 50)}}

    return ResourceUsageResponse(result)


def put_logs_auto_div(client, req, div=1):
    try:
        count = len(req.logitems)
        p1 = count // 2

        if div > 1 and p1 > 0:
            # divide req into multiple ones
            req1 = PutLogsRequest(project=req.project, logstore=req.logstore, topic=req.topic, source=req.source,
                                  logitems=req.logitems[:p1],
                                  hashKey=req.hashkey, compress=req.compress, logtags=req.logtags)
            req2 = PutLogsRequest(project=req.project, logstore=req.logstore, topic=req.topic, source=req.source,
                                  logitems=req.logitems[p1:],
                                  hashKey=req.hashkey, compress=req.compress, logtags=req.logtags)
            res = put_logs_auto_div(client, req1)
            return put_logs_auto_div(client, req2)
        else:
            return client.put_logs(req)
    except LogException as ex:
        if ex.get_error_code() == 'InvalidLogSize':
            return put_logs_auto_div(client, req, div=2)
        raise ex


def _transform_events_to_logstore(runner, events, to_client, to_project, to_logstore):
    count = removed = processed = failed = 0
    new_events = defaultdict(list)

    default_time = time.time()
    for src_event in events:
        count += 1
        new_event = runner(src_event)

        if new_event is None:
            removed += 1
            continue

        if not isinstance(new_event, (tuple, list)):
            new_event = (new_event, )

        for event in new_event:
            if not isinstance(event, dict):
                logger.error("transform_data: get unknown type of processed event: {0}".format(event))
                continue

            dt = int(event.get('__time__', default_time)) // 60  # group logs in same minute
            topic = ''
            source = ''
            if "__topic__" in event:
                topic = event['__topic__']
                del event["__topic__"]
            if "__source__" in event:
                source = event['__source__']
                del event["__source__"]

            new_events[(dt, topic, source)].append(event)

    for (dt, topic, source), contents in six.iteritems(new_events):

        items = []

        for content in contents:
            st = content.get("__time__", default_time)
            if "__time__" in content:
                del content['__time__']

            ct = list(six.iteritems(content))
            item = LogItem(st, ct)
            items.append(item)

        req = PutLogsRequest(project=to_project, logstore=to_logstore, topic=topic, source=source, logitems=items)
        res = put_logs_auto_div(to_client, req)
        processed += len(items)

    return count, removed, processed, failed


def transform_worker(from_client, from_project, from_logstore, shard_id, from_time, to_time,
                     config,
                     to_client, to_project, to_logstore, batch_size=None, compress=None,
                     ):
    next_cursor = "As from_time configured"
    try:
        runner = Runner(config)
        iter_data = from_client.pull_log(from_project, from_logstore, shard_id, from_time, to_time, batch_size=batch_size,
                                         compress=compress)

        count = removed = processed = failed = 0
        for s in iter_data:
            events = s.get_flatten_logs_json_auto()

            c, r, p, f = _transform_events_to_logstore(runner, events, to_client, to_project, to_logstore)
            count += c
            removed += r
            processed += p
            failed += f

            next_cursor = s.next_cursor
        return shard_id, count, removed, processed, failed
    except Exception as ex:
        logger.error("transform data failed: task info {0} failed to copy data to target, next cursor: {1} detail: {2}".
                     format(
            (from_project, from_logstore, shard_id, from_time, to_time, to_client, to_project, to_logstore),
            next_cursor, ex), exc_info=True)
        raise


class TransformDataConsumer(ConsumerProcessorBase):
    runner = None
    to_client = None
    to_project = None
    to_logstore = None

    def __init__(self, status_updator):
        super(TransformDataConsumer, self).__init__()
        self.count = self.removed = self.processed = self.failed = 0
        self.status_updator = status_updator

    @staticmethod
    def set_transform_options(config, to_client, to_project, to_logstore):
        TransformDataConsumer.runner = Runner(config)
        TransformDataConsumer.to_client = to_client
        TransformDataConsumer.to_project = to_project
        TransformDataConsumer.to_logstore = to_logstore

    def process(self, log_groups, check_point_tracker):
        logger.info("TransformDataConsumer::process: get log groups")

        logs = PullLogResponse.loggroups_to_flattern_list(log_groups)
        c, r, p, f = _transform_events_to_logstore(self.runner, logs, self.to_client, self.to_project, self.to_logstore)

        self.count += c
        self.removed += r
        self.processed += p
        self.failed += f

        # save check point
        self.save_checkpoint(check_point_tracker)

    def shutdown(self, check_point_tracker):
        super(TransformDataConsumer, self).shutdown(check_point_tracker)
        self.status_updator(self.shard_id, self.count, self.removed, self.processed, self.failed)


def transform_data(from_client, from_project, from_logstore, from_time,
                   to_time=None,
                   to_client=None, to_project=None, to_logstore=None,
                   shard_list=None,
                   config=None,
                   batch_size=None, compress=None,
                   cg_name=None, c_name=None,
                   cg_heartbeat_interval=None, cg_data_fetch_interval=None, cg_in_order=None,
                   cg_worker_pool_size=None
                   ):
    """
    transform data from one logstore to another one (could be the same or in different region), the time is log received time on server side.

    """
    if not config:
        logger.info("transform_data: config is not configured, use copy data by default.")
        return copy_data(from_client, from_project, from_logstore, from_time, to_time=to_time,
                         to_client=to_client, to_project=to_project, to_logstore=to_logstore,
                         shard_list=shard_list,
                         batch_size=batch_size, compress=compress)

    to_client = to_client or from_client

    # increase the timeout to 2 min at least
    from_client.timeout = max(from_client.timeout, 120)
    to_client.timeout = max(to_client.timeout, 120)
    to_project = to_project or from_project
    to_logstore = to_logstore or from_logstore

    if not cg_name:
        # batch mode
        to_time = to_time or "end"
        cpu_count = multiprocessing.cpu_count() * 2
        shards = from_client.list_shards(from_project, from_logstore).get_shards_info()
        current_shards = [str(shard['shardID']) for shard in shards]
        target_shards = _parse_shard_list(shard_list, current_shards)

        worker_size = min(cpu_count, len(target_shards))

        result = dict()
        total_count = 0
        total_removed = 0
        with ProcessPoolExecutor(max_workers=worker_size) as pool:
            futures = [pool.submit(transform_worker, from_client, from_project, from_logstore, shard,
                                   from_time, to_time, config,
                                   to_client, to_project, to_logstore,
                                   batch_size=batch_size, compress=compress)
                       for shard in target_shards]

            for future in as_completed(futures):
                if future.exception():
                    logger.error("get error when transforming data: {0}".format(future.exception()))
                else:
                    partition, count, removed, processed, failed = future.result()
                    total_count += count
                    total_removed += removed
                    if count:
                        result[partition] = {"total_count": count, "transformed":
                            processed, "removed": removed, "failed": failed}

        return LogResponse({}, {"total_count": total_count, "shards": result})

    else:
        # consumer group mode
        c_name = c_name or "transform_data_{0}".format(multiprocessing.current_process().pid)
        cg_heartbeat_interval = cg_heartbeat_interval or 20
        cg_data_fetch_interval = cg_data_fetch_interval or 2
        cg_in_order = False if cg_in_order is None else cg_in_order
        cg_worker_pool_size = cg_worker_pool_size or 3

        option = LogHubConfig(from_client._endpoint, from_client._accessKeyId, from_client._accessKey,
                              from_project, from_logstore, cg_name,
                              c_name, cursor_position=CursorPosition.SPECIAL_TIMER_CURSOR,
                              cursor_start_time=from_time,
                              cursor_end_time=to_time,
                              heartbeat_interval=cg_heartbeat_interval, data_fetch_interval=cg_data_fetch_interval,
                              in_order=cg_in_order,
                              worker_pool_size=cg_worker_pool_size)

        TransformDataConsumer.set_transform_options(config, to_client, to_project, to_logstore)

        result = {"total_count": 0, "shards": {}}
        l = RLock()

        def status_updator(shard_id, count=0, removed=0, processed=0, failed=0):
            logger.info("status update is called, shard: {0}, count: {1}, removed: {2}, processed: {3}, failed: {4}".format(shard_id, count, removed, processed, failed))

            with l:
                result["total_count"] += count
                if shard_id in result["shards"]:
                    data = result["shards"][shard_id]
                    result["shards"][shard_id] = {"total_count": data["total_count"] + count, "transformed": data["transformed"] + processed, "removed": data["removed"] + removed, "failed": data["failed"] + failed}
                else:
                    result["shards"][shard_id] = {"total_count": count, "transformed": processed, "removed": removed, "failed": failed}

        worker = ConsumerWorker(TransformDataConsumer, consumer_option=option, args=(status_updator, ) )
        worker.start()

        try:
            while worker.is_alive():
                worker.join(timeout=60)
            logger.info("transform_data: worker exit unexpected, try to shutdown it")
            worker.shutdown()
        except KeyboardInterrupt:
            logger.info("transform_data: *** try to exit **** ")
            print("try to stop transforming data.")
            worker.shutdown()
            worker.join(timeout=120)

        return LogResponse({}, result)



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.

from .logresponse import LogResponse
from .queriedlog import QueriedLog
from .logexception import LogException
import six


class GetContextLogsResponse(LogResponse):
    """ The response of the GetContextLog API from log.

    :type resp: dict
    :param resp: GetContextLogsResponse HTTP response body

    :type header: dict
    :param header: GetContextLogsResponse HTTP response header
    """

    def __init__(self, resp, header):
        LogResponse.__init__(self, header, resp)
        try:
            self.progress = resp["progress"]
            self.total_lines = resp["total_lines"]
            self.back_lines = resp["back_lines"]
            self.forward_lines = resp["forward_lines"]
            self.logs = []
            for data in resp["logs"]:
                contents = {}
                source = ""
                if "__source__" in data:
                    source = data['__source__']

                for key in six.iterkeys(data):
                    if key != '__time__' and key != '__source__':
                        contents[key] = data[key]
                self.logs.append(QueriedLog(data['__time__'], source, contents))
        except Exception as ex:
            raise LogException("InvalidResponse",
                               "Failed to parse GetContextLogResponse, \nheader: "
                               + str(header) + " \nBody:"
                               + str(resp) + " \nOther: " + str(ex),
                               resp_header=header,
                               resp_body=resp)

    def get_total_lines(self):
        """ Get log number from the response

        :return: int, log number
        """
        return self.total_lines

    def get_back_lines(self):
        """ Get log number backward from the response

        :return: int, log number
        """
        return self.back_lines

    def get_forward_lines(self):
        """ Get log number forward from the response

        :return: int, log number
        """
        return self.forward_lines

    def is_completed(self):
        """ Check if the get logs query is completed

        :return: bool, true if this logs query is completed
        """
        return self.progress == 'Complete'

    def get_logs(self):
        """ Get all logs from the response

        :return: QueriedLog list, all log data
        """
        return self.logs

    def log_print(self):
        print('GetContextLogsResponse:')
        print('headers:', self.get_all_headers())
        print('total_lines:', self.get_total_lines())
        print('back_lines:', self.get_back_lines())
        print('forward_lines:', self.get_forward_lines())
        print('progress:', self.progress)
        print('\nQueriedLog class:\n')
        for log in self.logs:
            log.log_print()
            print('\n')



from ..etl_util import u

__all__ = ['F_TAGS', 'F_META', 'F_TIME',
           'EMPTY', 'EXIST', 'NONE', 'NO_EMPTY', 'ANY', 'ALL']

ALL, ANY = True, True

F_TAGS = u'__tag__:.+'
F_META = u'__tag__:.+|__topic__|__source__'
F_TIME = u'__time__'


class EMPTY(object):
    def __init__(self, field):
        self.field = u(field)

    def __call__(self, event):
        return self.field not in event or not event[self.field]


class NONE(object):
    def __init__(self, field):
        self.field = u(field)

    def __call__(self, event):
        return self.field not in event


class EXIST(object):
    def __init__(self, field):
        self.field = u(field)

    def __call__(self, event):
        return self.field in event


class NO_EMPTY(object):
    def __init__(self, field):
        self.field = u(field)

    def __call__(self, event):
        return self.field in event and event[self.field]



import json
import six
from ..etl_util import u
from ..etl_util import get_re_full_match, get_set_mode_if_skip_fn
import logging

logger = logging.getLogger(__name__)


class trans_comp_base(object):
    @property
    def __name__(self):
        return str(type(self))

    @staticmethod
    def _n(v):
        """
        convert string to utf8 in Py2 or unicode in Py3
        :param v:
        :return:
        """
        if v is None:
            return ""

        if isinstance(v, (dict, list)):
            try:
                v = json.dumps(v)
            except Exception:
                pass
        elif six.PY2 and isinstance(v, six.text_type):
            v = v.encode('utf8', "ignore")
        elif six.PY3 and isinstance(v, six.binary_type):
            v = v.decode('utf8', "ignore")

        return str(v)

    @staticmethod
    def _u(d):
        """
        convert string, string container or unicode
        :param d:
        :return:
        """
        return u(d)


class trans_comp_check_mdoe_base(trans_comp_base):
    DEFAULT_KEYWORD_PTN = u'[\u4e00-\u9fa5\u0800-\u4e00a-zA-Z][\u4e00-\u9fa5\u0800-\u4e00\\w\\.\\-]*'
    SET_MODE = {
                "fill": get_set_mode_if_skip_fn(False, True, False),
                "add": get_set_mode_if_skip_fn(True, False, False),
                "overwrite": get_set_mode_if_skip_fn(False, False, False),
                "fill-auto": get_set_mode_if_skip_fn(False, True, True),
                "add-auto": get_set_mode_if_skip_fn(True, False, True),
                "overwrite-auto": get_set_mode_if_skip_fn(False, False, True)
                }
    DEFAULT_SET_MODE = 'fill-auto'

    def __init__(self, mode=None):
        super(trans_comp_check_mdoe_base, self).__init__()
        self.kw_ptn = get_re_full_match(self.DEFAULT_KEYWORD_PTN)
        self.skip_if = self.SET_MODE.get(mode, self.SET_MODE[self.DEFAULT_SET_MODE])

    def set(self, e, k, v, real_k=None, check_kw_name=False):
        if not check_kw_name or (check_kw_name and self.kw_ptn(k)):
            real_k = real_k or k
            if k and not self.skip_if(e, k, v):
                e[real_k] = v
                return True

        logger.debug("{1}: skip detected k-v due to current mode: {0}".format((k, v), type(self)))
        return False

    def sets(self, e, e_new, check_kw_name=False):
        has_update = False
        for k, v in six.iteritems(e_new):
            has_update = self.set(e, k, v, check_kw_name=check_kw_name) or has_update

        return has_update




from functools import wraps
import re
import copy
import six

DEFAULT_FN_CACHE_SIZE = 102400  # 100K

def cached(fn):
    @wraps(fn)
    def _wrapped(*args, **kwargs):
        sig = str(args) + str(kwargs)
        if not hasattr(fn, "__CACHE__") or len(fn.__CACHE__) > DEFAULT_FN_CACHE_SIZE:
            setattr(fn, "__CACHE__", {})

        if sig not in fn.__CACHE__:
            ret = fn(*args, **kwargs)
            fn.__CACHE__[sig] = ret
            return ret
        else:
            return fn.__CACHE__[sig]

    return _wrapped


def re_full_match(pattern, string, *args, **kwargs):
    is_not = isinstance(pattern, NOT)
    if six.PY2 and isinstance(pattern, six.binary_type):
        pattern = pattern.decode('utf8', 'ignore')
    if six.PY2 and isinstance(string, six.binary_type):
        string = string.decode('utf8', 'ignore')

    ret = None
    if hasattr(re, 'fullmatch'):
        ret = re.fullmatch(pattern, string, *args, **kwargs)
    else:
        m = re.match(pattern, string, *args, **kwargs)
        if m and m.span()[1] == len(string):
            ret = m

    return not ret if is_not else ret


# this function is used to bypass lambda trap
def bind_event_fn(fn, *args, **kwargs):
    @wraps(fn)
    def _real_fn(e):
        return fn(e, *args, **kwargs)

    return _real_fn


@cached
def get_re_full_match(pattern, flags=0):
    if six.PY2 and isinstance(pattern, six.binary_type):
        pattern = pattern.decode('utf8', 'ignore')

    p = re.compile(pattern, flags=flags)
    if hasattr(p, 'fullmatch'):  # normally it's Py3
        return p.fullmatch

    def ptn_full_match(string, *args, **kwargs):
        if six.PY2 and isinstance(string, six.binary_type):
            string = string.decode('utf8', 'ignore')

        m = p.match(string, *args, **kwargs)
        if m and m.span()[1] == len(string):
            return m

    return ptn_full_match


def process_event(event, fn_list, cp=True):
    if not len(fn_list):
        return event

    new_event = event
    if cp:
        new_event = copy.copy(event)

    if isinstance(new_event, dict):
        new_event = fn_list[0](new_event)
        return process_event(new_event, fn_list[1:], cp=False)
    elif isinstance(new_event, (tuple, list)):
        result = []
        for e in new_event:
            new_e = fn_list[0](e)
            ret = process_event(new_e, fn_list[1:], cp=False)
            if ret is None:
                continue

            if isinstance(ret, (tuple, list)):
                result.extend(ret)
            else:
                result.append(ret)

        if result:
            if len(result) == 1:
                return result[0]
            return result
        return None  # return None for empty list

    return new_event


def _is_event_list(event):
    if isinstance(event, (list, tuple)):
        for e in event:
            if e is not None and not isinstance(e, dict):
                break
        else:
            return True

    return False


def support_event_list_simple(fn):
    """
    enable __call__ to accept event_list.
    :param fn:
    :return:
    """
    @wraps(fn)
    def _wrapped(self, event, *args, **kwargs):
        if _is_event_list(event):
            result = []
            for e in event:
                ret = fn(self, e, *args, **kwargs)

                if ret is None:
                    continue

                if isinstance(ret, (tuple, list)):
                    result.extend(ret)
                else:
                    result.append(ret)

            if result:
                if len(result) == 1:
                    return result[0]
                return result
            return None
        else:
            return fn(self, event, *args, **kwargs)

    return _wrapped


def u(d):
    """
    convert string, string container or unicode
    :param d:
    :return:
    """
    if six.PY2:
        if isinstance(d, six.binary_type):
            return d.decode("utf8", "ignore")
        elif isinstance(d, list):
            return [u(x) for x in d]
        elif isinstance(d, tuple):
            return tuple(u(x) for x in d)
        elif isinstance(d, dict):
            return dict( (u(k), u(v)) for k, v in six.iteritems(d))

    return d


class NOT(object):
    def __new__(self, v):
        if isinstance(v, six.binary_type):
            return _NOT_B(v)
        elif isinstance(v, six.text_type):
            return _NOT_U(v)
        else:
            raise ValueError("NOT can only be used with string. ")


class _NOT_B(six.binary_type, NOT):
    def __init__(self, v):
        super(_NOT_B, self).__init__(v)
        self.v = v

    def decode(self, *args, **kwargs):
        return _NOT_B(self.v.decode(*args, **kwargs))


class _NOT_U(six.text_type, NOT):
    def __init__(self, v):
        if six.PY2:
            super(_NOT_U, self).__init__(v)
        self.v = v

    def encode(self, *args, **kwargs):
        return _NOT_B(self.v.encode(*args, **kwargs))


def get_set_mode_if_skip_fn(skip_if_src_exist, skip_if_src_not_empty, skip_if_value_empty):
    def if_skip(event, key, value):
        if skip_if_src_exist and key in event:
            return True
        if skip_if_src_not_empty and key in event and event[key]:
            return True
        if skip_if_value_empty and not value:
            return True

        return False
    return if_skip




#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.

import time
import copy


class LogItem(object):
    """ LogItem used to present a log, it contains log time and multiple
    key/value pairs to present the log contents.
    
    :type timestamp: int
    :param timestamp: time of the log item, the default time is the now time.
    
    :type contents: tuple(key-value) list
    :param contents: the data of the log item, including many (key,value) pairs. 
    """

    def __init__(self, timestamp=None, contents=None):
        self.timestamp = int(timestamp) if timestamp else int(time.time())
        self.contents = copy.deepcopy(contents) if contents else []

    def push_back(self, key, value):
        """ Append a key/value pair as a log content to the log
        
        :type key: string
        :param key: log content key
        
        :type value: string
        :param value: log content value
        """
        self.contents.append((key, value))

    def get_contents(self):
        """ Get log contents
        
        :return: tuple(key-value) list, log contents.
        """
        return self.contents

    def set_contents(self, contents):
        """ Set log contents
        
        :type contents: tuple(key-value) list
        :param contents: log contents (key-value pair list)
        """
        self.contents = copy.deepcopy(contents)

    def get_time(self):
        """ Get log time
        
        :return: int, log time
        """
        return self.timestamp

    def set_time(self, timestamp):
        """ Set log time
        
        :type timestamp: int
        :param timestamp: log time
        """
        self.timestamp = int(timestamp)

    def log_print(self):
        print('time', self.timestamp)
        print('contents', self.contents)



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.


import json
import re

from .util import split_and_strip


class IndexLogstoreMappings(object):

    def __init__(self, index_lst=None, logstore_index_mappings=None):
        self.index_logstore_dct = {}
        self.logstore_indexes_dct = {}

        if not index_lst:
            return
        self.index_logstore_dct = {index: index for index in index_lst}

        logstore_index_dct = {}
        if logstore_index_mappings:
            logstore_index_dct = json.loads(logstore_index_mappings)
        self._update_dicts(logstore_index_dct)

    def _update_dicts(self, logstore_index_dct):
        all_indexes = self.index_logstore_dct.keys()

        for k, v in logstore_index_dct.items():
            indexes = split_and_strip(v, ",")
            index_lst = []
            for pattern in indexes:
                match_index_lst = self._get_match_indexes(pattern, all_indexes)
                index_lst.extend(match_index_lst)
                for index in match_index_lst:
                    if index in self.index_logstore_dct and \
                            self.index_logstore_dct[index] in logstore_index_dct:
                        raise Exception(
                            "index '%s' belongs to '%s' and '%s'" % (index, k, self.index_logstore_dct[index]))
                    self.index_logstore_dct[index] = k
            if index_lst:
                self.logstore_indexes_dct[k] = index_lst

        for v in self.index_logstore_dct.values():
            if v not in self.logstore_indexes_dct:
                self.logstore_indexes_dct[v] = [v]

    def get_logstore(self, index):
        if index in self.index_logstore_dct:
            return self.index_logstore_dct[index]
        return None

    def get_all_logstores(self):
        return self.logstore_indexes_dct.keys()

    def get_indexes(self, logstore):
        if logstore in self.logstore_indexes_dct:
            return self.logstore_indexes_dct[logstore]
        return []

    def get_all_indexes(self):
        return self.index_logstore_dct.keys()

    @classmethod
    def _get_match_indexes(cls, pattern, index_lst):
        if not pattern or not index_lst:
            return []
        if "*" in pattern:
            regex = re.compile(pattern.replace("*", ".*"))
            match_index_lst = [index for index in index_lst if re.match(regex, index)]
        else:
            match_index_lst = [index for index in index_lst if pattern == index]
        return match_index_lst



import six, json
from .logclient_operator import list_more
from .logexception import LogException
from .pluralize import pluralize
from .common_response import *

DEFAULT_MAX_LIST_PAGING_SIZE = 500


def create_entity(entity_name, logstore_level=None, root_resource=None):
    def do_create(self, project, detail, resource_prefix=None):
        params = {}
        headers = {"x-log-bodyrawsize": '0', "Content-Type": "application/json"}

        if hasattr(detail, 'to_json'):
            detail = detail.to_json()
            body_str = six.b(json.dumps(detail))
        elif isinstance(detail, six.binary_type):
            body_str = detail
        elif isinstance(detail, six.text_type):
            body_str = detail.encode('utf8')
        else:
            body_str = six.b(json.dumps(detail))

        resource_path = (root_resource and root_resource.rstrip('/')) or "/" + pluralize(entity_name)
        if resource_prefix:
            resource_path = resource_prefix + resource_path
        (resp, header) = self._send("POST", project, body_str, resource_path, params, headers)
        return GetEntityResponse(header, resp)

    if not logstore_level:
        def fn(self, project, detail):
            """ Create {entity_title}.
            Unsuccessful opertaion will cause an LogException.

            :type project: string
            :param project: project name

            :type detail: dict/string
            :param detail: json string

            :return: CreateEntityResponse

            :raise: LogException
            """
            return do_create(self, project, detail)
    else:
        def fn(self, project, logstore, detail):
            """ Create {entity_title}.
            Unsuccessful opertaion will cause an LogException.

            :type project: string
            :param project: project name

            :type logstore: string
            :param logstore: logstore name

            :type detail: dict/string
            :param detail: json string

            :return: CreateEntityResponse

            :raise: LogException
            """
            resource_prefix = "/logstores/" + logstore
            return do_create(self, project, detail, resource_prefix)

    fn.__name__ = 'create_' + entity_name
    fn.__doc__ = fn.__doc__.format(entity_title=entity_name.title())
    return fn


def get_entity(entity_name, logstore_level=None, root_resource=None):
    if not logstore_level:
        def fn(self, project, entity):
            """Get {entity_title}.
            Unsuccessful opertaion will cause an LogException.

            :type project: string
            :param project: project name

            :type entity: string
            :param entity: {entity_name} name

            :return: GetEntityResponse

            :raise: LogException
            """
            resource_path = ((root_resource and root_resource.rstrip('/')) or ('/' + pluralize(entity_name))) + '/' + entity

            (resp, header) = self._send("GET", project, None, resource_path, dict(), dict())
            return GetEntityResponse(header, resp)
    else:
        def fn(self, project, logstore, entity):
            """Get {entity_title}.
            Unsuccessful opertaion will cause an LogException.

            :type project: string
            :param project: project name

            :type logstore: string
            :param logstore: logstore name

            :type entity: string
            :param entity: {entity_name} name

            :return: GetEntityResponse

            :raise: LogException
            """
            resource_path = ((root_resource and root_resource.rstrip('/')) or ('/' + pluralize(entity_name))) + '/' + entity
            resource_path = "/logstores/" + logstore + resource_path

            (resp, header) = self._send("GET", project, None, resource_path, dict(), dict())
            return GetEntityResponse(header, resp)

    fn.__name__ = 'get_' + entity_name
    fn.__doc__ = fn.__doc__.format(entity_name=entity_name, entity_title=entity_name.title())

    return fn


def delete_entity(entity_name, logstore_level=None, root_resource=None):
    if not logstore_level:
        def fn(self, project, entity):
            """Delete {entity_title}.
            Unsuccessful opertaion will cause an LogException.

            :type project: string
            :param project: project name

            :type entity: string
            :param entity: {entity_name} name

            :return: DeleteEntityResponse

            :raise: LogException
            """
            resource_path = ((root_resource and root_resource.rstrip('/')) or ('/' + pluralize(entity_name))) + '/' + entity
            (resp, header) = self._send("DELETE", project, None, resource_path, dict(), dict())
            return DeleteEntityResponse(header, resp)
    else:
        def fn(self, project, logstore, entity):
            """Delete {entity_title}.
            Unsuccessful opertaion will cause an LogException.

            :type project: string
            :param project: project name

            :type logstore: string
            :param logstore: logstore name

            :type entity: string
            :param entity: {entity_name} name

            :return: DeleteEntityResponse

            :raise: LogException
            """
            resource_path = ((root_resource and root_resource.rstrip('/')) or ('/' + pluralize(entity_name))) + '/' + entity
            resource_path = "/logstores/" + logstore + resource_path
            (resp, header) = self._send("DELETE", project, None, resource_path, dict(), dict())
            return DeleteEntityResponse(header, resp)

    fn.__name__ = 'delete_' + entity_name
    fn.__doc__ = fn.__doc__.format(entity_name=entity_name, entity_title=entity_name.title())

    return fn


def list_entity(entity_name, logstore_level=None, root_resource=None, max_batch_size=DEFAULT_MAX_LIST_PAGING_SIZE, entities_key=None, raw_resource_name=None):
    def do_list(self, project, resource_path, offset=0, size=100):
        headers = {}
        params = {}
        resource_name = raw_resource_name if raw_resource_name else pluralize(entity_name)
        params['offset'] = str(offset)
        params['size'] = str(size)
        (resp, header) = self._send("GET", project, None, resource_path, params, headers)
        return ListEntityResponse(header, resp, resource_name=resource_name, entities_key=entities_key)

    if not logstore_level:
        def fn(self, project, offset=0, size=100):
            """ list the {entity_title}, get first 100 items by default
            Unsuccessful opertaion will cause an LogException.

            :type project: string
            :param project: the Project name

            :type offset: int
            :param offset: the offset of all the matched names

            :type size: int
            :param size: the max return names count, -1 means all

            :return: ListLogStoreResponse

            :raise: LogException
            """

            # need to use extended method to get more
            if int(size) == -1 or int(size) > max_batch_size:
                return list_more(fn, int(offset), int(size), max_batch_size, self, project)

            resource_path = (root_resource and root_resource.rstrip('/')) or "/" + pluralize(entity_name)
            return do_list(self, project, resource_path, offset=offset, size=size)

    else:
        def fn(self, project, logstore, offset=0, size=100):
            """ list the {entity_title}, get first 100 items by default
            Unsuccessful opertaion will cause an LogException.

            :type project: string
            :param project: the Project name

            :type logstore: string
            :param logstore: the logstore name

            :type offset: int
            :param offset: the offset of all the matched names

            :type size: int
            :param size: the max return names count, -1 means all

            :return: ListLogStoreResponse

            :raise: LogException
            """

            # need to use extended method to get more
            if int(size) == -1 or int(size) > max_batch_size:
                return list_more(fn, int(offset), int(size), max_batch_size, self, project, logstore)

            resource_path = (root_resource and root_resource.rstrip('/')) or "/" + pluralize(entity_name)
            resource_path = "/logstores/" + logstore + resource_path
            return do_list(self, project, resource_path, offset=offset, size=size)

    fn.__name__ = 'list_' + entity_name
    fn.__doc__ = fn.__doc__.format(entity_title=entity_name.title())

    return fn


def update_entity(entity_name, logstore_level=None, name_field=None, root_resource=None):
    def do_update(self, project, detail, resource_prefix=None):
        params = {}
        headers = {}

        # parse entity value
        entity = None
        if hasattr(detail, 'to_json'):
            detail = detail.to_json()
            body_str = six.b(json.dumps(detail))
            entity = detail.get(name_field or 'name', '')
        elif isinstance(detail, six.binary_type):
            body_str = detail
        elif isinstance(detail, six.text_type):
            body_str = detail.encode('utf8')
        else:
            body_str = six.b(json.dumps(detail))
            entity = detail.get(name_field or 'name', '')

        if entity is None:
            entity = json.loads(body_str).get(name_field, '')

        assert entity, LogException('InvalidParameter', 'unknown entity name "{0}" in "{1}"'.format(name_field, detail))
        resource_path = ((root_resource and root_resource.rstrip('/')) or ('/' + pluralize(entity_name))) + '/' + entity
        if resource_prefix:
            resource_path = resource_prefix + resource_path

        headers['Content-Type'] = 'application/json'
        headers['x-log-bodyrawsize'] = str(len(body_str))
        (resp, headers) = self._send("PUT", project, body_str, resource_path, params, headers)
        return UpdateEntityResponse(headers, resp)

    if not logstore_level:
        def fn(self, project, detail):
            """ Update {entity_title}.
            Unsuccessful opertaion will cause an LogException.

            :type project: string
            :param project: project name

            :type detail: dict/string
            :param detail: json string

            :return: UpdateEntityResponse

            :raise: LogException
            """
            return do_update(self, project, detail)
    else:
        def fn(self, project, logstore, detail):
            """ Update {entity_title}.
            Unsuccessful opertaion will cause an LogException.

            :type project: string
            :param project: project name

            :type logstore: string
            :param logstore: logstore name

            :type detail: dict/string
            :param detail: json string

            :return: UpdateEntityResponse

            :raise: LogException
            """
            resource_prefix = "/logstores/" + logstore
            return do_update(self, project, detail, resource_prefix)

    fn.__name__ = 'update_' + entity_name
    fn.__doc__ = fn.__doc__.format(entity_title=entity_name.title())

    return fn


def make_lcrud_methods(obj, entity_name, logstore_level=None, name_field=None, root_resource=None, entities_key=None, raw_resource_name=None):
    setattr(obj, 'list_' + entity_name, list_entity(entity_name, logstore_level=logstore_level, root_resource=root_resource, entities_key=entities_key, raw_resource_name=raw_resource_name))
    setattr(obj, 'get_' + entity_name, get_entity(entity_name, logstore_level=logstore_level, root_resource=root_resource))
    setattr(obj, 'delete_' + entity_name, delete_entity(entity_name, logstore_level=logstore_level, root_resource=root_resource))
    setattr(obj, 'update_' + entity_name, update_entity(entity_name, logstore_level=logstore_level, root_resource=root_resource, name_field=name_field))
    setattr(obj, 'create_' + entity_name, create_entity(entity_name, logstore_level=logstore_level, root_resource=root_resource))




from .trans_base import trans_comp_check_mdoe_base
import csv
import six
import re
try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable
from ..exceptions import SettingError
import logging

__all__ = ['trans_comp_csv', 'trans_comp_tsv', 'trans_comp_psv']

logger = logging.getLogger(__name__)


class trans_comp_csv(trans_comp_check_mdoe_base):
    p_csv_sep = re.compile(r'\s*,\s*')
    DEFAULT_SEP = ','
    DEFAULT_QUOTE = '"'

    def __init__(self, config, sep=None, quote=None, lstrip=None, restrict=None, mode=None):
        super(trans_comp_csv, self).__init__(mode=mode)
        config = self._u(config)
        if isinstance(config, (six.text_type, six.binary_type)):
            self.keys = self.p_csv_sep.split(config)
        elif isinstance(config, Iterable):
            self.keys = list(config)
        else:
            raise SettingError(settings=config)

        self.sep = sep or self.DEFAULT_SEP
        self.quote = quote or self.DEFAULT_QUOTE
        self.lstrip = True if lstrip is None else lstrip
        self.restrict = False if restrict is None else restrict

    def __call__(self, event, inpt):
        inpt = self._u(inpt)
        if inpt in event:
            data = event[inpt].split("\n")
            ret = list(csv.reader(data, skipinitialspace=self.lstrip, delimiter=self.sep, quotechar=self.quote))[0]
            if self.restrict and len(ret) != len(self.keys):
                logger.warning(
                    u"event {0} field {1} contains different count of fields as expected key {2} actual {3}".format(
                        event, inpt, self.keys, ret))
                return event

            new_event = dict(zip(self.keys, ret))
            self.sets(event, new_event)
        else:
            logger.warning(u"field {0} doesn't exist in event {1}, skip it".format(inpt, event))

        return event


class trans_comp_tsv(trans_comp_csv):
    DEFAULT_SEP = '\t'


class trans_comp_psv(trans_comp_csv):
    DEFAULT_SEP = '|'
